# -*- coding: utf-8 -*-
"""
Cleans data prior to doing matching. Called by the matching.py file.
"""
# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

import math
import pandas as pd
import numpy as np

from . import early_stops

def read_files(input_data, holdout_data):
    """Both options can be either df or csv files and are parsed here.

    Input:
        input_data: The matching data as string filename or df
        holdout_data: The holdout data as string filename, df, fraction, bool

    Return:
        input_data: dataframe
        holdout_data: dataframe
    """

    # Read the input data
    if type(input_data) == pd.core.frame.DataFrame:
        df_input = input_data

    elif not input_data:
        raise Exception("Need to specify either csv file name or pandas data "\
                        "frame in parameter 'input_data'")
    else:
        df_input = pd.read_csv(input_data)

    # Now read the holdout data
    if type(holdout_data) == pd.core.frame.DataFrame:
        df_holdout = holdout_data
    elif (type(holdout_data) == float and holdout_data <= 1.0
          and holdout_data > 0.0):
        df_holdout = df_input.sample(frac=holdout_data)
    elif not holdout_data:
        df_holdout = df_input.sample(frac=0.1) # default if it's not provided is df.
    else:
        df_holdout = pd.read_csv(holdout_data)

    df_input.columns = map(str, df_input.columns)
    df_holdout.columns = map(str, df_holdout.columns)

    return df_input, df_holdout

def check_stops(stop_unmatched_c, early_stop_un_c_frac, stop_unmatched_t,
                early_stop_un_t_frac, early_stop_pe, early_stop_pe_frac,
                early_stop_iterations):
    """Check the parameters passed to DAME/FLAME relating to early stopping"""

    early_stops_obj = early_stops.EarlyStops()

    if (early_stop_un_t_frac > 1.0 or early_stop_un_t_frac < 0.0):
        raise Exception('The value provided for the early stopping critera '\
                        'of proportion of unmatched treatment units needs to '\
                        'be between 0.0 and 1.0')

    if (early_stop_un_c_frac > 1.0 or early_stop_un_c_frac < 0.0):
        raise Exception('The value provided for the early stopping critera '\
                        'of proportion of unmatched control units needs to '\
                        'be between 0.0 and 1.0')

    if early_stop_pe:
        early_stop_pe = early_stop_pe_frac
    if (early_stop_pe_frac > 1.0 or early_stop_pe_frac < 0.0):
        raise Exception('The value provided for the early stopping critera of'\
                        ' PE needs to be between 0.0 and 1.0')

    if (type(early_stop_iterations) != int and early_stop_iterations):
        raise Exception('The value provided for early_stop_iteration needs '\
                        'to be an integer number of iterations, or False if '\
                        'not stopping early based on the number of iterations')

    # Put all of those parameters into the object to return
    early_stops_obj.unmatched_c = stop_unmatched_c
    early_stops_obj.unmatched_t = stop_unmatched_t
    early_stops_obj.un_c_frac = early_stop_un_c_frac
    early_stops_obj.un_t_frac = early_stop_un_t_frac
    early_stops_obj.pe = early_stop_pe
    early_stops_obj.iterations = early_stop_iterations

    return early_stops_obj

def check_parameters(adaptive_weights, df_holdout, df_input, alpha, FLAME,
                     weight_array=[], C=0.0, verbose=0):
    '''
    This function processes the parameters that were passed to DAME/FLAME
    that aren't directly the input file or related to stop_criteria.
    '''

    # Checks on the weight array...if the weight array needs to exist
    if not adaptive_weights:

        if type(weight_array) != list:
            raise Exception('Invalid input error. A weight array of type'\
                            'array needs to be provided when the'\
                            'parameter adaptive_weights == True')

        # Confirm that weight array has the right number of values in it
        # Subtracting 2 because one col is the treatment and one is outcome.
        if len(weight_array) != (len(df_input.columns)-2):
            raise Exception('Invalid input error. Weight array size not equal'\
                            ' to number of columns in dataframe')

        # Confirm that weights in weight vector add to 1.
        if abs(sum(weight_array) - 1.0) >= 0.001:
            # I do this weird operation instead of seeing if it equals one
            # to avoid floatig point addition errors that can occur.
            raise Exception('Invalid input error. Weight array values must '\
                            'sum to 1.0')

    else:
        # make sure that the alpha is valid if it's a ridge regression.
        if (adaptive_weights in ["ridge", "ridgeCV"] and alpha < 0.0):
            raise Exception('Invalid input error. The alpha needs to be '\
                            'positive for ridge regressions.')

        # make sure that adaptive_weights is a valid value.
        if (adaptive_weights not in ["ridge", "decisiontree", "ridgeCV", "decisiontreeCV"]):
            raise Exception("Invalid input error. The acceptable values for "\
                            "the adaptive_weights parameter are 'ridge', "\
                            "'decisiontree', 'decisiontreeCV', or 'ridgeCV'. Additionally, "\
                            "adaptive-weights may be 'False' along "\
                            "with a weight array")


        # make sure the two dfs have the same number of columns first:
        if len(df_input.columns) != len(df_holdout.columns):
            raise Exception('Invalid input error. The holdout and main '\
                            'dataset must have the same number of columns')

        # make sure that the holdout columns match the df columns.
        if set(df_holdout.columns) != set(df_input.columns):
            # they don't match
            raise Exception('Invalid input error. The holdout and main '\
                            'dataset must have the same columns')

    if FLAME:
        if C < 0.0:
            raise Exception('The C, or the hyperparameter to trade-off between'\
                           ' balancing factor and predictive error must be '\
                           ' nonnegative. ')


def replace_unique_large(df, treatment_column_name, outcome_column_name,
                         missing_indicator):
    ''' (helper)
    This function replaces missing values from the df with unique large values
    could possibly clean this up later
    '''
    # now we replace all of the missing_indicators with unique large vals
    # that are larger than max_val.

    df = df.replace(missing_indicator, np.nan)

    for col in df.columns:
        if col not in [treatment_column_name, outcome_column_name]:
            max_val = df[col].max()
            for item_num in df.index.values:
                if math.isnan(df[col][item_num]):
                    df.loc[item_num, col] = max_val + 1
                    max_val += 1

    cols = list(df.columns)
    cols.remove(outcome_column_name)
    df[cols] = df[cols].astype('int64')

    return df


def drop_missing(df, missing_indicator):
    '''
    helper, this function drops rows that have missing_indicator in any of the cols
    '''

    if type(missing_indicator) != str and math.isnan(missing_indicator) == True:
        # either the missing indicator is already NaN and we just drop those rows
        df = df.dropna().copy()
    else:
        # but if its not NaN, switch missing_indicator with nan and then drop
        df = df.replace(missing_indicator, np.nan)
        df = df.dropna().copy()
    return df


def check_missings(df_input, df_holdout, missing_indicator, missing_data_replace,
                   missing_holdout_replace, missing_holdout_imputations,
                   missing_data_imputations, treatment_column_name,
                   outcome_column_name, adaptive_weights):
    '''
    This function deals with all the missing data related stuff
    '''
    # No missing data allowed in the fixed weights version.
    # This is because having a weight array for pre-determined weights wouldn't
    # make sense...how would a user have even predetermined the weights?
    # If this is changed later, pay attention to MICE case (multiple weight
    # arrays needed?) and also sorting columns in replace unique large.
    if (missing_data_replace != 0 and not adaptive_weights):
        raise Exception('Invalid input error. We do not support missing data '\
                        'handing in the fixed weights version of algorithms')

    mice_on_matching = False
    mice_on_holdout = False
    if (missing_data_replace == 0 and df_input.isnull().values.any()):
        print('There is missing data in this dataset. The default missing '\
              'data handling is being done, so we are not matching on '\
              'any missing values in the matching set')
        missing_data_replace = 2

    if missing_data_replace == 1:
        df_input = drop_missing(df_input, missing_indicator)

    if missing_data_replace == 2:
        # so replacing with large unique values will only work if columns
        # are in order!!

        df_input = replace_unique_large(df_input, treatment_column_name,
                                  outcome_column_name, missing_indicator)

        # Reorder if they're not in order:
        df_input = df_input.loc[:, df_input.max().sort_values(ascending=True).index]

    if missing_data_replace == 3:
        # this means do mice but only if theres something actually missing.
        df_input = df_input.replace(missing_indicator, np.nan)
        if df_input.isnull().values.any():
            mice_on_matching = missing_data_imputations

    if missing_holdout_replace == 0 and df_holdout.isnull().values.any():
        print('There is missing data in this dataset. The default missing '\
              'data handling is being done, so we are running MICE on 10 '\
              'imputed holdout datasets')
        missing_holdout_replace = 2

    if missing_holdout_replace == 1:
        df_holdout = drop_missing(df_holdout, missing_indicator)

    if missing_holdout_replace == 2:
        # this means do mice.
        df_holdout = df_holdout.replace(missing_indicator, np.nan)
        # but if there is actually nothing missing in the dataset, then dont
        # need to do this.
        if df_holdout.isnull().values.any():
            mice_on_holdout = missing_holdout_imputations

    # converts float inputs to ints
    if not mice_on_matching:
        try:
            cols = list(df_input.columns)
            cols.remove(outcome_column_name)
            df_input[cols] = df_input[cols].astype('int64')
        except:
            raise Exception('Invalid input error on matching dataset. Ensure '\
                            'all inputs asides from the outcome column are '\
                            'integers, and if missing values exist, ensure '\
                            'they are handled.')
    if not mice_on_holdout:
        try:
            cols = list(df_holdout.columns)
            cols.remove(outcome_column_name)
            df_holdout[cols] = df_holdout[cols].astype('int64')
        except:
            raise Exception('Invalid input error on holdout dataset. Ensure '\
                            'all inputs asides from the outcome column are '\
                            'integers, and if missing values exist, ensure '\
                            'they are handled.')

    return df_input, df_holdout, mice_on_matching, mice_on_holdout

def process_input_file(df, treatment_column_name, outcome_column_name):
    '''
    This function processes the parameters passed to DAME/FLAME that are
    directly the input file.

    '''

    # Confirm that the treatment column name exists.
    if treatment_column_name not in df.columns:
        raise Exception('Invalid input error. Treatment column name does not'\
                        ' exist')

    # Confirm that the outcome column name exists.
    if outcome_column_name not in df.columns:
        raise Exception('Invalid input error. Outcome column name does not'\
                        ' exist')

    # column only has 0s and 1s.
    if set(df[treatment_column_name].unique()) != {0, 1}:
        raise Exception('Invalid input error. All rows in the treatment '\
                        'column must have either a 0 or a 1 value.')

    return df



# -*- coding: utf-8 -*-
"""The main file for the FLAME algorithm"""

# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

import pandas as pd
import numpy as np

from . import grouped_mr
from . import dame_algorithm
from . import flame_dame_helpers

def decide_drop(all_covs, consider_dropping, prev_drop, df_all,
                treatment_column_name, outcome_column_name, df_holdout_array,
                adaptive_weights, alpha_given, df_unmatched, return_matches,
                C, weight_array):
    """
    This is a helper function, where we decide which covar to drop next

    Args:
        all_covs (array): Array of covar column names, not including treatment
            and outcome columns.
        consider_dropping (set): Covariate column names that have not yet
            been dropped in a previous iteration
        prev_drop (set): Covariate column names that have been dropped
            in a previous iteration

    """

    # This is where we decide who to drop, and also compute the pe
    # value that gets outputted in the list described in readme.
    best_drop = 0
    best_mq = float("-inf")
    best_return_matches = 0
    best_matched_rows = 0
    best_bf = 0
    best_pe = 0
    best_units_in_g = 0

    if not adaptive_weights:
        # find the covariate that can be dropped with the minimum value in
        # the weight array
        min_covar_weight = 1
        best_drop = 0
        for poss_drop in consider_dropping:
            index_in_all_covs = all_covs.index(poss_drop)
            covar_weight = weight_array[index_in_all_covs]
            if covar_weight < min_covar_weight:
                min_covar_weight = covar_weight
                best_drop = poss_drop

        all_covs = set(all_covs)
        covs_match_on = all_covs.difference([best_drop]).difference(prev_drop)
        covs_match_on = list(covs_match_on)

        # need to make sure we don't edit the mutable dataframes, then do match
        df_all_temp = df_all.copy(deep=True)
        return_matches_temp = return_matches.copy(deep=True)
        matched_rows, return_matches, units_in_g = grouped_mr.algo2_GroupedMR(
            df_all_temp, df_unmatched, covs_match_on, all_covs,
            treatment_column_name, outcome_column_name, return_matches_temp)

        # find the BF for this covariate set's match.
        BF = flame_dame_helpers.compute_bf(matched_rows,
                                           treatment_column_name, df_unmatched)

        return best_drop, 0, matched_rows, return_matches, BF, units_in_g

    else:
        for poss_drop in consider_dropping:
            # S is the set of covars we drop. We try dropping each one
            s = prev_drop.union([poss_drop])

            PE = flame_dame_helpers.find_pe_for_covar_set(
                df_holdout_array, treatment_column_name, outcome_column_name, s,
                adaptive_weights, alpha_given)

            # error check. PE can be float(0), but not denote error
            if not PE and type(PE) == bool:
                return False, False, False, False, False

            # The dropping criteria for FLAME is max MQ
            # MQ = C * BF - PE

            all_covs = set(all_covs)
            covs_match_on = all_covs.difference([poss_drop]).difference(prev_drop)
            covs_match_on = list(covs_match_on)

            # need to make sure we don't edit the mutable dataframes, then do match
            df_all_temp = df_all.copy(deep=True)
            return_matches_temp = return_matches.copy(deep=True)
            matched_rows, return_matches_temp, units_in_g = grouped_mr.algo2_GroupedMR(
                df_all_temp, df_unmatched, covs_match_on, all_covs,
                treatment_column_name, outcome_column_name, return_matches_temp)

            # find the BF for this covariate set's match.
            BF = flame_dame_helpers.compute_bf(
                matched_rows, treatment_column_name, df_unmatched)

            # Use the largest MQ as the covariate set to drop.
            MQ = C * BF - PE
            if MQ > best_mq:
                best_mq = MQ
                best_pe = PE
                best_bf = BF
                best_drop = poss_drop
                best_return_matches = return_matches_temp
                best_matched_rows = matched_rows
                best_units_in_g = units_in_g

        return best_drop, best_pe, best_matched_rows, best_return_matches, best_bf, best_units_in_g

def flame_generic(df_all, treatment_column_name, weight_array,
                  outcome_column_name, adaptive_weights, alpha, df_holdout,
                  repeats, want_pe, verbose, want_bf, missing_holdout_replace,
                  early_stops, pre_dame, C):
    '''
    All variables are the same as dame algorithm 1 except for:
    pre_dame(False, integer): Indicates whether the algorithm will move to
    DAME and after integer number of iterations.
    '''

    # Initialize variables. These are all moving/temporary throughout algo
    all_covs = df_all.columns.tolist()
    all_covs.remove(treatment_column_name)
    all_covs.remove(outcome_column_name)
    df_unmatched = df_all.copy(deep=True)

    # The items getting returned
    return_pe = [] # list of predictive errors,
    return_bf = []
    MG_units = [] # list of unit ids for each matched group
    # weights indicates the number of times each unit appears in a group
    weights = pd.DataFrame(np.zeros(shape=(len(df_all.index), 1)),
                           columns=['weights'],
                           index=df_all.index)

    return_matches = pd.DataFrame(columns=all_covs, index=df_all.index)

    # Initialize variables used in checking stopping criteria
    orig_len_df_all = len(df_all) # Need this bc of case where repeats=False
    orig_tot_treated = df_all[treatment_column_name].sum()

    # As an initial step, we attempt to match on all covariates

    covs_match_on = all_covs
    matched_rows, return_matches, units_in_g = grouped_mr.algo2_GroupedMR(
        df_all, df_unmatched, covs_match_on, all_covs, treatment_column_name,
        outcome_column_name, return_matches)

    if (len(units_in_g)) != 0:
        # add the newly matched groups to MG_units, which tracks units in groups
        MG_units = MG_units + units_in_g
        # update unit weights for all units which appear in the new groups
        # flatten to 1 list, then add occurrences of unique units
        flat_units_in_g = np.concatenate(units_in_g).ravel()
        unique_units, occurrences = np.unique(flat_units_in_g, return_counts=True)
        for index in range(len(unique_units)):
            weights['weights'][unique_units[index]] += occurrences[index]

    # Now remove the matched units
    df_unmatched.drop(matched_rows.index, inplace=True)

    if not repeats:
        df_all = df_unmatched

    # set up all the extra dfs if needed
    if missing_holdout_replace:
        # now df_holdout is actually an array of imputed datasets
        df_holdout_array = flame_dame_helpers.create_mice_dfs(
            df_holdout, missing_holdout_replace, outcome_column_name)
    else:
        # df_holdout_array exists regardless, just size 1 and equal to itself
        # if not doing mice.
        df_holdout_array = list()
        df_holdout_array.append(df_holdout)


    h = 1 # The iteration number

    if verbose == 3:
        flame_dame_helpers.verbose_output(h, len(MG_units),
            df_unmatched[treatment_column_name].sum(), len(df_unmatched),
            orig_len_df_all, orig_tot_treated, 0, orig_len_df_all, set())

    prev_iter_num_unmatched = len(df_unmatched) # this is for output progress
    consider_dropping = set(i for i in all_covs)
    prev_dropped = set()

    # Here, we begin the iterative dropping procedure of FLAME
    while True:
        # see if any stopping criteria have been met        
        if (flame_dame_helpers.stop_iterating(early_stops, df_unmatched,
                                              repeats, treatment_column_name,
                                              orig_len_df_all, h,
                                              orig_tot_treated,
                                              consider_dropping)):
            break

        new_drop, pe, matched_rows, return_matches, bf, units_in_g = decide_drop(all_covs,
            consider_dropping, prev_dropped, df_all, treatment_column_name,
            outcome_column_name, df_holdout_array, adaptive_weights, alpha,
            df_unmatched, return_matches, C, weight_array)

        # Check for error in above step:
        if not new_drop:
            raise Exception("There may have been an error in your choice of "\
                            "machine learning algorithm used to choose the "\
                            "covariate to drop. For help, please reach on "\
                            "github to the team. ")

        if (len(units_in_g)) != 0:
        # add the newly matched groups to MG_units, which tracks units in groups
            MG_units = MG_units + units_in_g
            # update unit weights for all units which appear in the new groups
            # flatten to 1 list, then add occurrences of unique units
            flat_units_in_g = np.concatenate(units_in_g).ravel()
            unique_units, occurrences = np.unique(flat_units_in_g, return_counts=True)
            for index in range(len(unique_units)):
                weights['weights'][unique_units[index]] += occurrences[index]

        return_pe.append(pe)

        if want_bf:
            # if we need to track the bf, do so.
            return_bf.append(bf)

        if early_stops.pe:
            if pe >= early_stops.pe:
                print((orig_len_df_all - len(df_unmatched)), "units matched. "\
                        "We stopped matching with a pe of ", pe)
                break

        # Update covariate groups for future iterations
        consider_dropping = consider_dropping.difference([new_drop])
        prev_dropped.add(new_drop)

        # Remove matches.
        df_unmatched = df_unmatched.drop(matched_rows.index, errors='ignore')

        if not repeats:
            df_all = df_unmatched

        h += 1

        # End of iter. Prints output based on verbose.
        if verbose == 1:
            print("Iteration number: ", h)
        if ((verbose == 2 and (h%10 == 0)) or verbose == 3):

            flame_dame_helpers.verbose_output(h, len(MG_units),
                df_unmatched[treatment_column_name].sum(), len(df_unmatched),
                orig_len_df_all, orig_tot_treated, pe, prev_iter_num_unmatched,
                new_drop)

            if want_bf:
                print("\tBalancing Factor of this iteration: ", bf)

        # Do we switch to DAME?
        if (pre_dame and pre_dame <= h):

            # drop the columns that have already been matched on
            for i in prev_dropped:
                df_all = df_all.loc[:, df_all.columns.drop(i)]
                df_holdout = df_holdout.loc[:, df_holdout.columns.drop(i)]


            # call dame algorithm
            print((orig_len_df_all - len(df_unmatched)), "units matched. "\
                  "Moving to DAME algorithm")
            return_matches_dame = dame_algorithm.algo1(
                df_all, treatment_column_name, weight_array,
                outcome_column_name, adaptive_weights, alpha, df_holdout,
                repeats, want_pe, verbose, want_bf, missing_holdout_replace,
                early_stops)

            # when dame is done, we
            # return the matches we made here, plus the matches made in dame.

            # but first, make sure anything not matched isn't in the df:
            return_matches = return_matches.dropna(axis=0) #drop rows with nan
            return_matches = return_matches.join(weights)
            return_package = [return_matches, MG_units]
            if want_pe:
                return_package.append(return_pe)
            if want_bf:
                return_package.append(return_bf)
            return_package.append(return_matches_dame)
            return return_package


        # end loop.

    return_matches = return_matches.dropna(axis=0) #drop rows with nan
    return_package = [return_matches]

    # append weights and MGs to return package
    return_package[0] = return_package[0].join(weights)
    return_package.append(MG_units)

    if want_pe:
        return_package.append(return_pe)
    if want_bf:
        return_package.append(return_bf)

    return return_package



# -*- coding: utf-8 -*-

"""Does the bit matching algo from "Fast Large Scale..."(Wang, et al)"""

# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

from operator import itemgetter
import numpy as np
from . import flame_group_by

def algo2_GroupedMR(df_all, df_unmatched, covs_match_on, all_covs, treatment_column_name,
                    outcome_column_name, return_groups):
    '''
    Input:
        df_all: The dataframe of all of the data
        df_unmatched: The dataframe ofjust hte unmatched data, a subset of df
        covs_match_on (array): List of strs with name of columns of df.
            A subset of indexes of all covariates.
        all_covs (array): list of all covariates.
    Output:
        matched_rows: newly matched units using covs indexed by Js. Type df
        return_groups: The df of unit id and covar values matched on, with '*"
            for the irrelevant ones.
        all_units_in_g: list of unit ids for all matched groups created

    '''
    # Find max of columns and make sure list of columns and list of maximums
    # is sorted from least to greatest

    # This is a list of tuples (max_of_column, column_name)
    covs_max_tuples = [(max(df_all[x])+1, x) for x in covs_match_on]
    covs_max_tuples = sorted(covs_max_tuples, key=itemgetter(0))

    # Now covs_match_on is a list of covars and covs_max_list is their maximums
    covs_max_list, covs_match_on = map(list, zip(*covs_max_tuples))


    # Form groups on D by exact matching on Js.

    df_all_without_outcome = df_all.drop([outcome_column_name], axis=1)

    matched_units, bi = flame_group_by.match_ng(df_all_without_outcome,
                                                covs_match_on, covs_max_list,
                                                treatment_column_name)
    # Find newly matched units and their main matched groups.

    # These are the rows of the ones that have been matched:
    matched_rows = df_all.loc[matched_units, :].copy()
    matched_rows['b_i'] = bi

    # These are the unique values in the bi col. length = number of groups
    unique_matched_row_vals = np.unique(bi)

    # Each element in this list represents a matched group and contains all of
    # the unit ids belonging to that particular group
    all_units_in_g = []
    for bi_val in unique_matched_row_vals:
        # type "int64index", ~ list, all of the unit_numbers in a matched group.
        units_in_g = matched_rows.index[matched_rows['b_i'] == bi_val]

        # Which of the units of this new group haven't been matched yet?
        # unique_matched is a subset of units in the matched group, just the
        # ones for whom this is their main matched group.
        newly_matched = [i for i in units_in_g if i in df_unmatched.index]
        # Only need to proceed to fill in the return table if someone's MMG found.
        if len(newly_matched) != 0:

            all_units_in_g.append(list(units_in_g))

            # Now, we figure out: What does the group look like? eg [1,2,*,1]
            temp_row_in_group = matched_rows.loc[units_in_g[0]]
            # ^ that line arbitrarily chooses the first row that has the bi_val so
            # the first row of that group.
            group_covs = []
            for col in return_groups.columns:
                if col in covs_match_on:
                    group_covs.append(temp_row_in_group[col])
                else:
                    group_covs.append('*')
            # add that group to the newly matched units to our new dataframe

            return_groups.loc[newly_matched, :] = group_covs

            # other idea for this step of algorithm:
            # store the bi in a column with df_all and also a column for "pair" with another
            # persons unit id.
            # don't update that when someone gets added to an auxiliary matched group
            # then at the end, iterate through it and create the nicely formatted output.

    return matched_rows, return_groups, all_units_in_g



# -*- coding: utf-8 -*-
"""Generate data for examples"""

# author: Neha Gupta, Tianyu Wang, Duke University; Awa Dieng, Yameng Liu
# Copyright Duke University 2020
# License: MIT

import pandas as pd
import numpy as np

def generate_uniform_given_importance(num_control=1000, num_treated=1000,
                                      num_cov=4, min_val=0,
                                      max_val=3, covar_importance=[4, 3, 2, 1],
                                      bi_mean=2, bi_stdev=1):
    """
    This generates data according to the discrete uniform distribution
    """
    x_c = np.random.randint(min_val, max_val, size=(num_control, num_cov))
    x_t = np.random.randint(min_val, max_val, size=(num_treated, num_cov))

    y_c = np.dot(x_c, np.array(covar_importance)) # y for control group

    # this is beta
    treatment_eff_coef = np.random.normal(bi_mean, bi_stdev, size=num_cov)
    treatment_effect = np.dot(x_t, treatment_eff_coef) # this is beta*x

    # yc is just the 1st term of the below summation. Thus, CATT is the 2nd term
    y_t = np.dot(x_t, np.array(covar_importance)) + treatment_effect
    true_catt = treatment_effect

    df1 = pd.DataFrame(x_c, columns=range(num_cov))
    df1['outcome'] = y_c
    df1['treated'] = 0

    df2 = pd.DataFrame(x_t, columns=range(num_cov))
    df2['outcome'] = y_t
    df2['treated'] = 1
    data_frame = pd.concat([df2, df1])

    data_frame = data_frame.reset_index()
    data_frame = data_frame.drop(['index'], axis=1)

    return data_frame, true_catt

def generate_binomial_given_importance(num_control=1000, num_treated=1000,
                                       num_cov=5, bernoulli_param=0.5,
                                       bi_mean=2, bi_stdev=1,
                                       covar_importance=[4, 3, 2, 1, 0.01]):
    '''
    This function generates data where the covariates exponentially decay with
    importance. The x's are all binary.
    '''
     # data for control group
    x_c = np.random.binomial(1, bernoulli_param, size=(num_control, num_cov))

    # data for treated group
    x_t = np.random.binomial(1, bernoulli_param, size=(num_treated, num_cov))

    y_c = np.dot(x_c, np.array(covar_importance)) # y for control group

    # this is beta
    treatment_eff_coef = np.random.normal(bi_mean, bi_stdev, size=num_cov)
    treatment_effect = np.dot(x_t, treatment_eff_coef) # this is beta*x

    # yc is just the 1st term of the below summation. Thus, CATT is the 2nd term
    y_t = np.dot(x_t, np.array(covar_importance)) + treatment_effect
    true_catt = treatment_effect

    df1 = pd.DataFrame(x_c, columns=range(num_cov))
    df1['outcome'] = y_c
    df1['treated'] = 0

    df2 = pd.DataFrame(x_t, columns=range(num_cov))
    df2['outcome'] = y_t
    df2['treated'] = 1
    data_frame = pd.concat([df2, df1])

    data_frame = data_frame.reset_index()
    data_frame = data_frame.drop(['index'], axis=1)

    return data_frame, true_catt


def generate_binomial_decay_importance(num_control=1000, num_treated=1000,
                                       num_cov=5, bernoulli_param=0.5,
                                       bi_mean=2, bi_stdev=1):
    '''
    This function generates data where the covariates exponentially decay with
    importance. The x's are all binary.
    '''
    # data for control group
    x_c = np.random.binomial(1, bernoulli_param, size=(num_control, num_cov))

    # data for treated group
    x_t = np.random.binomial(1, bernoulli_param, size=(num_treated, num_cov))

    dense_bs = [64*((1/4)**(i+1)) for i in range(num_cov)]

    y_c = np.dot(x_c, np.array(dense_bs)) # y for control group

    # this is beta
    treatment_eff_coef = np.random.normal(bi_mean, bi_stdev, size=num_cov)
    treatment_effect = np.dot(x_t, treatment_eff_coef) # this is beta*x

    # yc is just the 1st term of the below summation. Thus, CATT is the 2nd term
    y_t = np.dot(x_t, np.array(dense_bs)) + treatment_effect
    true_catt = treatment_effect

    df1 = pd.DataFrame(x_c, columns=range(num_cov))
    df1['outcome'] = y_c
    df1['treated'] = 0

    df2 = pd.DataFrame(x_t, columns=range(num_cov))
    df2['outcome'] = y_t
    df2['treated'] = 1
    data_frame = pd.concat([df2, df1])

    data_frame = data_frame.reset_index()
    data_frame = data_frame.drop(['index'], axis=1)

    return data_frame, true_catt



# -*- coding: utf-8 -*-
"""Algorithm to generate sets in "Dynamic Almost Matching" (Liu, et al)"""

# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

import itertools

def algo3GenerateNewActiveSets(newly_dropped, prev_processed):
    """This function does Algorithm 3 in the paper.

    Args:
        newly_dropped: A newly dropped set of size k. Stored as type frozenset.
            This is s in the paper.
        prev_processed: the set of previously processed sets. Stored as set of
            frozensets. This is delta in the paper.

    Returns:
        Z, the new active sets. Stored as set of frozensets.

    """
    new_active_sets = set() # this is Z in the paper
    size_newly_dropped = len(list(newly_dropped)) # this is k in the paper
    set_newly_dropped = set(newly_dropped)

    # Step 3: delta_k is a set of frozensets.
    # it contains all subsets of delta that are size k, and s.
    subsets_size_k = set()
    for fs_prevprocessed in prev_processed:
        if len(fs_prevprocessed) == size_newly_dropped:
            subsets_size_k.add(fs_prevprocessed)
    subsets_size_k.add(newly_dropped)
    delta_k = subsets_size_k

    # Step 4: rho is all the covariates contained in sets in delta_k
    rho = set()
    for tup in delta_k:
        for i in tup:
            rho.add(i)

    # Step 5: s_e is the support of covariate e in delta_k, for covars in rho
    # possible time efficiency exploration: combine this with above loop, 
    # calculating rho in one step
    s_e = dict()
    for covar_e in rho:
        s_e[covar_e] = 0
        for tup in delta_k:
            if covar_e in tup:
                s_e[covar_e] += 1

    # Step 6: omega is all the covariates not in s that have enough support
    # .items() iterates through key,val pairs
    # so dict_covars_eno is the dictionary of all covariates that have enough support
    dict_covars_eno = dict((key, val) for key, val in s_e.items() if val >= size_newly_dropped)
    omega = set(dict_covars_eno.keys())
    omega = omega.difference(set_newly_dropped)


    # Step 7: do all covariates in S have enough support in delta_k?
    for covar_e, support_e in s_e.items():
        if covar_e in newly_dropped and support_e < size_newly_dropped:
            return new_active_sets

    # Step 8
    for alpha in omega:
        # Step 9
        r = set_newly_dropped.union(set([alpha]))

        # Step 10: Get all subsets of size_newly_dropped in r, check if belong in delta_k

        # Have to convert these to sets and frozen sets before doing the search
        delta_k = set(frozenset(i) for i in delta_k)
        subsets_size_k = set(list(itertools.combinations(r, size_newly_dropped)))
        subsets_size_k = set(frozenset(i) for i in subsets_size_k)

        allin = True
        for subset in subsets_size_k:
            if subset not in delta_k:
                allin = False
                break

        if allin:
            # Step 11: Add r to Z
            new_active_sets.add(frozenset(r))

    return new_active_sets



# -*- coding: utf-8 -*-
""" Group-by bit match for FLAME algorithm, from "Fast Large..."(Wang etal)"""

# author: Neha Gupta, Tianyu Wang, Duke University
# Copyright Duke University 2020
# License: MIT

import numpy as np

def match_ng(df, covs, covs_max_list, treatment_indicator_col='treated'):
    '''
    This is the match_ng function
    '''
    # this function takes a dataframe, a set of covariates to match on,
    # the treatment indicator column and the matched indicator column.
    # it returns the array indicating whether each unit is matched (the first return value),
    # and a list of indices for the matched units (the second return value)

    arr_slice_wo_t = df[covs].values # the covariates values as a matrix

    # the covariate values together with the treatment indicator as a matrix
    arr_slice_w_t = df[covs + [treatment_indicator_col]].values

    # matrix multiplication, get a unique number for each unit
    b_i = np.dot(arr_slice_wo_t, np.array([covs_max_list[i]**(i)
                                           for i in range(len(covs_max_list))]))

    # matrix multiplication, get a unique number for each unit with treatment indicator
    b_i_plus = np.dot(arr_slice_w_t, np.array([covs_max_list[i]**(i+1)
                                               for i in range(len(covs_max_list))] + [1]))

    # count how many times each number appears
    _, unqtags_wo_t, c_i = np.unique(b_i, return_inverse=True,
                                     return_counts=True)

    # count how many times each number appears (with treatment indicator)
    _, unqtags_w_t, c_i_plus = np.unique(b_i_plus, return_inverse=True,
                                         return_counts=True)

    # a unit is matched if and only if the counts don't agree
    match_indicator = ~(c_i_plus[unqtags_w_t] == c_i[unqtags_wo_t])
    return match_indicator, b_i[match_indicator]



# -*- coding: utf-8 -*-
"""For organizing all of the early stopping criteria options"""
# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

class EarlyStops:
    '''
    Class to hold all early stopping criteria as decided by user
    '''
    def __init__(self):
        '''
        Constructor, defaults all of the possible early stopping criteria
        '''
        # create the members
        self.unmatched_c = False
        self.unmatched_t = False
        self.un_c_frac = 0.1
        self.un_t_frac = 0.1
        self.pe = 0.1
        self.iterations = False



# -*- coding: utf-8 -*-
""" Find Treatment Effect Estimators, and Matched Groups, of matched units.
"""

# author: Neha Gupta, Tommy Howell, Duke University
# Copyright Duke University 2020
# License: MIT

import numpy as np
from .. import matching

def validate_matching_obj(matching_object):
    """ Check matching_object's type and that .fit(), .predict() done"""

    if matching.MatchParent not in type(matching_object).__bases__:
        raise Exception("The matching_object input parameter needs to be "\
                        "of type DAME or FLAME")
    if (not hasattr(matching_object, 'input_data')):
        raise Exception("This function can be only called after a match has "\
                       "been formed using the .fit() and .predict() functions")

def MG(matching_object, unit_ids, output_style=1, mice_iter=0):
    '''
    This function returns the main matched groups for all specified unit
    indices

    Args:
        matching_object (DAME or FLAME object)
        unit_ids (int, list): units for which MG will be returned

    Returns:
        MMGs: list of datraframes or singular dataframe containing the units
            in the main matched groups for the specified units

    '''
    # Accept int or list for unit_id
    if type(unit_ids) is int:
        unit_ids = [unit_ids]

    validate_matching_obj(matching_object)

    if (matching_object.missing_data_replace != 3):
        array_mgs = matching_object.units_per_group
        df_matched_units = matching_object.df_units_and_covars_matched
    else:
        array_mgs = matching_object.units_per_group[mice_iter]
        df_matched_units = matching_object.df_units_and_covars_matched[mice_iter]


    main_matched_groups = []

    # Now we recover MMG
    for unit in unit_ids:
        if unit in df_matched_units.index:
            # Iterate through all matched groups
            for group in array_mgs:
                # The first group to contain the specified unit is the MMG
                if unit in group:
                    new_group = matching_object.input_data.loc[group]
                    my_series = df_matched_units.loc[unit]
                    if output_style == 1 and "*" in my_series.unique():
                        # Insert asterisks for unused covariates
                        star_cols = my_series[my_series == "*"].index
                        for col in star_cols:
                            new_group[[col]] = ['*'] * len(new_group.index)
                    main_matched_groups.append(new_group)
                    break
        # Warn user if a unit has no matches
        else:
            main_matched_groups.append(np.nan)
            print('Unit ' + str(unit) + ' does not have any matches')
    # Format output
    if len(main_matched_groups) == 1:
        main_matched_groups = main_matched_groups[0]
    return main_matched_groups


def CATE(matching_object, unit_ids, mice_iter=0):
    '''
    This function returns the CATEs for all specified unit indices

    Args:
        matching_object (DAME or FLAME object)
        unit_ids (int, list): units for which MG will be returned

    Returns:
        CATEs: list of floats or singular float containing the CATEs
            of the main matched groups for the specified units

    '''

    # Accept int or list for unit_id
    if type(unit_ids) is int:
        unit_ids = [unit_ids]

    validate_matching_obj(matching_object)


    if (matching_object.missing_data_replace != 3):
        arr_matched_groups = matching_object.units_per_group        
        df_matched_units = matching_object.df_units_and_covars_matched
    else:
        arr_matched_groups = matching_object.units_per_group[mice_iter]
        df_matched_units = matching_object.df_units_and_covars_matched[mice_iter]

    # Recover CATEs
    cates = []
    for unit in unit_ids:
        if unit in df_matched_units.index:
            for group in arr_matched_groups:
                # The first group to contain the specified unit is the MMG
                if unit in group:
                    df_mmg = matching_object.input_data.loc[group, 
                        [matching_object.treatment_column_name,
                         matching_object.outcome_column_name]]
                    break
            # Assuming an MMG has been found, compute CATE for that group
            treated = df_mmg.loc[df_mmg[matching_object.treatment_column_name] == 1]
            control = df_mmg.loc[df_mmg[matching_object.treatment_column_name] == 0]
            avg_treated = sum(treated[matching_object.outcome_column_name])/len(treated.index)
            avg_control = sum(control[matching_object.outcome_column_name])/len(control.index)
            cates.append(avg_treated - avg_control)
        # Warn user that unit has no matches
        else:
            cates.append(np.nan)
            if matching_object.verbose != 0:
                print('Unit ' + str(unit) + " does not have any matches, so " \
                      "can't find the CATE")

    # Format output
    if len(cates) == 1:
        cates = cates[0]
    return cates

def ATE(matching_object, mice_iter=0):
    '''
    This function returns the ATE for the matching data

    Args:
        matching_object (DAME or FLAME object)
    Returns:
        ATE: the average treatment effect for the matching data

    '''

    validate_matching_obj(matching_object)

    if matching_object.missing_data_replace != 3:
        arr_matched_groups = matching_object.units_per_group
        num_groups_per_unit = matching_object.groups_per_unit
    else:
        arr_matched_groups = matching_object.units_per_group[mice_iter]
        num_groups_per_unit = matching_object.groups_per_unit[mice_iter]

    # Recover CATEs
    cates = [0] * len(arr_matched_groups) # this will be a CATE for each matched group
    for group_id in range(len(arr_matched_groups)):
        group_data = matching_object.input_data.loc[arr_matched_groups[group_id],
                                                    [matching_object.treatment_column_name,
                                                     matching_object.outcome_column_name]]
        treated = group_data.loc[group_data[matching_object.treatment_column_name] == 1]
        control = group_data.loc[group_data[matching_object.treatment_column_name] == 0]
        avg_treated = sum(treated[matching_object.outcome_column_name]) / len(treated.index)
        avg_control = sum(control[matching_object.outcome_column_name]) / len(control.index)
        cates[group_id] = avg_treated - avg_control

    # Compute ATE
    weight_sum = 0
    weighted_cate_sum = 0
    for group_id in range(len(arr_matched_groups)):
        matched_group_weight = 0
        for unit in arr_matched_groups[group_id]:
            matched_group_weight += num_groups_per_unit[unit]
        weight_sum += matched_group_weight
        weighted_cate_sum += matched_group_weight * cates[group_id]

    return weighted_cate_sum/weight_sum



def ATT(matching_object, mice_iter=0):
    '''
    This function returns the ATT for the matching data using
    balancing estimation

    Args:
        matching_object (DAME or FLAME object)
    Returns:
        ATT: the average treatment effect on the treated for the matching data
    '''

    validate_matching_obj(matching_object)

    if matching_object.missing_data_replace != 3:
        num_groups_per_unit = matching_object.groups_per_unit
        matched_df = matching_object.input_data.loc[
            matching_object.df_units_and_covars_matched.index]
    else:
        num_groups_per_unit = matching_object.groups_per_unit[mice_iter]
        matched_df = matching_object.input_data.loc[
            matching_object.df_units_and_covars_matched[mice_iter].index]

    treated = matched_df.loc[matched_df[matching_object.treatment_column_name] == 1]
    control = matched_df.loc[matched_df[matching_object.treatment_column_name] == 0]
    # Compute ATT
    avg_treated = sum(treated[matching_object.outcome_column_name])/len(treated.index)
    # Stores weights for all control units
    control_weights = []
    for unit in control.index:
        if unit in num_groups_per_unit.index:
            control_weights.append(num_groups_per_unit[unit])
        else:
            # Unmatched units have a weight of 0
            control_weights.append(0)
    control_weight_sum = sum(control_weights)
    avg_control = sum(control[matching_object.outcome_column_name] * control_weights)/control_weight_sum
    return avg_treated - avg_control



# -*- coding: utf-8 -*-
""" The main place where the DAME algorithm happens"""
# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

import numpy as np
import pandas as pd
from . import grouped_mr
from . import generate_new_active_sets
from . import flame_dame_helpers



def decide_drop(all_covs, active_covar_sets, weights, adaptive_weights,
                treatment_column_name, outcome_column_name, df_holdout,
                alpha_given):
    """ This is a helper function to Algorithm 1 in the paper.

    Args:
        all_covs: This is an array of just the cov column names.
            Not including treat/outcome
        active_covar_sets: A set of frozensets, representing all the active
            covar sets
        weights: This is the weight array provided by the user
        adaptive_weights: This is the T/F provided by the user indicating
            whether to run ridge regression to decide who to drop.
        treatment_column_name (str): name of treatment column in df
        outcome_column_name (str): name of outcome column in df
        df_holdout: The cleaned, user-provided dataframe with all rows/columns.
            There are no changes made to this throughout the code. Used only in
            testing/training for adaptive_weights version.
    """
    curr_covar_set = set()
    best_pe = float("inf")
    if not adaptive_weights:
        # We iterate through all active covariate sets and find the total
        # weight of each . For each possible covariate set, temp_weight counts
        # the total weight of the covs that are going to get used in the match,
        # or the ones *not* in that possible cov set.
        max_weight = 0
        for s in active_covar_sets: # s is a set to consider dropping
            temp_weight = 0
            for cov_index in range(len(all_covs)):  # iter through all covars
                if all_covs[cov_index] not in s:
                    # if an item not in s, add weight. finding impact of drop s
                    temp_weight += weights[cov_index]
            if temp_weight >= max_weight:
                max_weight = temp_weight
                curr_covar_set = s # This is the items we will drop, that will
                # not get used in the match.
        best_pe = max_weight

    else:
        # Iterate through all of the active_covar_sets and drop one at a time,
        # and drop the one with the highest match quality score
        for s in active_covar_sets:
            # S is the frozenset of covars we drop. We try dropping each one
            PE = flame_dame_helpers.find_pe_for_covar_set(df_holdout,
                                                          treatment_column_name,
                                                          outcome_column_name, s, adaptive_weights,
                                                          alpha_given)
            # error check. PE can be float(0), but not denote error
            if not PE and type(PE) == bool:
                return False, False

            # Use the smallest PE as the covariate set to drop.
            if PE < best_pe:
                best_pe = PE
                curr_covar_set = s

    return curr_covar_set, best_pe


def algo1(df_all, treatment_column_name="T", weight_array=[],
          outcome_column_name="outcome", adaptive_weights=False, alpha=0.1,
          df_holdout="", repeats=True, want_pe=False, verbose=0,
          want_bf=False, missing_holdout_replace=False, early_stops=False):
    """This function does Algorithm 1 in the paper.

    Args:
        df_all: The cleaned, user-provided dataframe with all rows/columns.
            There are no changes made to this throughout the code when the
            parameter "repeats" is True. In the DAME paper, this is called 'D'.
        treatment_column_name: As provided by the user, this indicates the name
            of the column that contains the binary indicator for whether each
            row is a treatment group or not.
        weight_array: As provided by the user, array of weights of all covariates
            that are in df_all.
        outcome_column_name: As provided by the user, this indicates the name
            of the column that contains the outcome values.
        adaptive_weights: Provided by the user, this is true if decide to drop
            weights based on a ridge regression on hold-out training set
            or false (default) if decide to drop weights
            based on the weights given in the weight_array
        alpha (float): for ridge regression.
        df_holdout: The array of cleaned, user-provided dataframe with all
            rows/columns. There are no changes made to this throughout the
            code. Used only in testing/training for adaptive_weights version.
            Size 1 with 1 array if not doing MICE on holdout dataset.
        repeats (bool): Provided by user, whether or not values for whom a MMG
            has been found can be used again and placed in an auxiliary group.
        want_pe (bool): Whether or not we want predictive error of each match
        want_bf (bool): Whether to compute and output the balancing factor of
            each group.
        missing_holdout_replace (bool/float): Default false. If int, the number of
            imputations that MICE needs to do on the holdout dataset, which
            has NaNs in it that need to be replaced.
        early_stops (type EarlyStop): This is all of the possible stop criteria

    Returns:
        return_df: df of units with the column values of their main matched
            group, with "*"s in place for the columns not in their MMG;
            includes a unit weights column which indicates the number of times
            each unit was matched
        MG_units: list of unit ids for every matched group
    """

    # Initialize variables. These are all moving/temporary throughout algo
    all_covs = df_all.columns.tolist()
    all_covs.remove(treatment_column_name) # This is J in the paper
    all_covs.remove(outcome_column_name)
    df_unmatched = df_all.copy(deep=True) # This is df_h in the paper

    # Initialize return values
    return_pe = []
    return_bf = []
    MG_units = [] # list of unit ids for each matched group
    # weights indicates the number of times each unit appears in a group
    weights = pd.DataFrame(np.zeros(shape=(len(df_all.index), 1)),
                           columns=['weights'],
                           index=df_all.index)
    return_matches = pd.DataFrame(columns=all_covs, index=df_all.index)

    # Initialize variables used in checking stopping criteria
    orig_len_df_all = len(df_all) # Need this bc of case where repeats=False
    orig_tot_treated = df_all[treatment_column_name].sum()

    # As an initial step, we attempt to match on all covariates

    covs_match_on = all_covs
    matched_rows, return_matches, units_in_g = grouped_mr.algo2_GroupedMR(
        df_all, df_unmatched, covs_match_on, all_covs, treatment_column_name,
        outcome_column_name, return_matches)

    if (len(units_in_g)) != 0:
        # add the newly matched groups to MG_units, which tracks units in groups
        MG_units = MG_units + units_in_g
        # update unit weights for all units which appear in the new groups
        # flatten to 1 list, then add occurrences of unique units
        flat_units_in_g = np.concatenate(units_in_g).ravel()
        unique_units, occurrences = np.unique(flat_units_in_g, return_counts=True)
        for index in range(len(unique_units)):
            weights['weights'][unique_units[index]] += occurrences[index]

    # Now remove the matched units
    df_unmatched.drop(matched_rows.index, inplace=True)

    if not repeats:
        df_all = df_unmatched

    # set up all the extra dfs if needed
    if missing_holdout_replace:
        # now df_holdout is actually an array of imputed datasets
        df_holdout = flame_dame_helpers.create_mice_dfs(
            df_holdout, missing_holdout_replace, outcome_column_name)
    else:
        # df_holdout is type array regardless, just size 1 and equal to itself
        # if not doing mice.
        x = list()
        x.append(df_holdout)
        df_holdout = x
    # Here we initializing variables for the iterative portion of the code.
    # active_covar_sets indicates the sets elibible to be dropped. In the
    # paper, this is lambda_h. curr_covar_sets is the covariates chosen to be
    # dropped. In the paper, this is s*h. processed_covar_sets is the already
    # processed sets from previous iterations. In the paper, it's delta_h.

    active_covar_sets = set(frozenset([i]) for i in all_covs)
    processed_covar_sets = set()

    h = 1 # The iteration number

    if verbose == 3:
        flame_dame_helpers.verbose_output(h, len(MG_units),
                                          df_unmatched[treatment_column_name].sum(),
                                          len(df_unmatched), orig_len_df_all,
                                          orig_tot_treated, 0,
                                          orig_len_df_all, set())

    prev_iter_num_unmatched = len(df_unmatched) # this is for output progress

    # Here, we begin the iterative dropping procedure of DAME
    while True:

        # see if any stopping criteria have been met
        if (flame_dame_helpers.stop_iterating(early_stops, df_unmatched,
                                              repeats, treatment_column_name,
                                              orig_len_df_all, h,
                                              orig_tot_treated,
                                              active_covar_sets)):
            break

        # We find curr_covar_set, the best covariate set to drop.
        curr_covar_set, pe = decide_drop(
            all_covs, active_covar_sets, weight_array, adaptive_weights,
            treatment_column_name, outcome_column_name, df_holdout, alpha)

        # Check for error in above step:
        if not curr_covar_set:
            print((orig_len_df_all - len(df_unmatched)), "units matched. "\
                  "We stopped when the holdout set was not large enough or "\
                  "there was nothing left to match")
            break

        return_pe.append(pe)

        covs_match_on = list(set(all_covs)-curr_covar_set)

        matched_rows, return_matches, units_in_g = grouped_mr.algo2_GroupedMR(
            df_all, df_unmatched, covs_match_on, all_covs, treatment_column_name,
            outcome_column_name, return_matches)

        if (len(units_in_g)) != 0:
            # add the newly matched groups to MG_units, which tracks units in groups
            MG_units = MG_units + units_in_g
            # update unit weights for all units which appear in the new groups
            # flatten to 1 list, then add occurrences of unique units
            flat_units_in_g = np.concatenate(units_in_g).ravel()
            unique_units, occurrences = np.unique(flat_units_in_g, return_counts=True)
            for index in range(len(unique_units)):
                weights['weights'][unique_units[index]] += occurrences[index]


        # It's probably slow to compute this if people don't want it, so will
        # want to add this, I think.
        if want_bf:
            # compute balancing factor
            mg_treated = matched_rows[treatment_column_name].sum()
            mg_control = len(matched_rows) - mg_treated
            available_treated = df_unmatched[treatment_column_name].sum()
            available_control = len(df_unmatched) - available_treated
            if (available_treated != 0 and available_control != 0):
                bf = mg_treated/available_treated + mg_control/available_control
            else:
                bf = np.nan
            return_bf.append(bf)

        if early_stops.pe:
            if pe >= early_stops.pe:
                print((orig_len_df_all - len(df_unmatched)), "units matched. "\
                        "We stopped matching with a pe of ", pe)
                break


        # Generate new active sets
        Z_h = generate_new_active_sets.algo3GenerateNewActiveSets(curr_covar_set,
                                                                  processed_covar_sets)

        # Remove curr_covar_set from the set of active sets
        active_covar_sets = active_covar_sets.difference([curr_covar_set])

        # Update the set of active sets
        active_covar_sets = active_covar_sets.union(Z_h)

        # Update the set of already processed covariate-sets. This works bc
        # processed_covar_sets is type set, but curr_covar_set is type frozenset
        processed_covar_sets.add(curr_covar_set)

        # Remove matches.
        df_unmatched = df_unmatched.drop(matched_rows.index, errors='ignore')

        if not repeats:
            df_all = df_unmatched

        h += 1
        # End of iter. Decide what to print to user depending on verbose var.
        if verbose == 1:
            print("Iteration number: ", h)
        if ((verbose == 2 and (h%10 == 0)) or verbose == 3):

            flame_dame_helpers.verbose_output(h, len(MG_units),
                                              df_unmatched[treatment_column_name].sum(),
                                              len(df_unmatched),
                                              orig_len_df_all,
                                              orig_tot_treated, pe,
                                              prev_iter_num_unmatched,
                                              curr_covar_set)
            if want_bf:
                print("Balancing factor of this iteration: ", bf)

            prev_iter_num_unmatched = len(df_unmatched)

    # end loop.
    return_matches = return_matches.dropna(axis=0) # drop rows with nan, dont return unmatched stuff
    return_package = [return_matches]

    return_package[0] = return_package[0].join(weights)
    return_package.append(MG_units)

    # the optional returns
    if want_pe:
        return_package.append(return_pe)
    if want_bf:
        return_package.append(return_bf)


    return return_package



# -*- coding: utf-8 -*-
"""Testing file"""
# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

from dame_flame import matching
from dame_flame.utils.data import *
from dame_flame.utils.post_processing import *
import unittest
import pandas as pd
import os
import sys


def check_statistics(model,unit_id = 1):
    ATE_ = ATE(model)
    ATT_ = ATT(model)
    MG_ = MG(model,unit_id)
    CATE_ = CATE(model,unit_id)
    
    if len(model.units_per_group) == 0:
        return True
    if len(model.groups_per_unit) == 0:
        return True
    if type(ATE_) == np.nan:
        print("ATE: " + str(ATE_))
        return True
    if type(ATT_) == np.nan:
        print("ATT:" + str(ATT_))
        return True
    return False

class TestFlame(unittest.TestCase):
    '''
    This file tests the overall FLAME algorithm against results that were
    confirmed to be correct by comparing to the FLAME R package.
    '''
    
    def test_large_C_repeats_F(self):
        
        df_path = os.path.join((os.path.dirname(__file__)), 'basicTestData.csv')
        df = pd.read_csv(df_path)
    
        holdout_path = os.path.join((os.path.dirname(__file__)), 'basicHoldoutData.csv')
        holdout = pd.read_csv(holdout_path)
        model = matching.FLAME(repeats=False, verbose=1)
        model.fit(holdout_data=holdout)
        algo_output = model.predict(df, C=100000)
        
        result_path = os.path.join((os.path.dirname(__file__)), 'basicResultData.csv')
        result = pd.read_csv(result_path, index_col="Unnamed: 0")
        
        dfs_equal = 1
        try:
            for index in result.index:
                for col in result.columns:
                    if (result.loc[index, col] != "*" and
                        algo_output.loc[index, col] != "*" and
                        int(result.loc[index, col]) != int(algo_output.loc[index, col])):
                            print("index, col", index, col)
                            dfs_equal = 0
                            break
        except (KeyError, ValueError):
            # We would hit this block if theres a key error, so df columns
            # are not equal or have different units, or weird entry in df, (string)
            dfs_equal = 0
        
        self.assertEqual(1, dfs_equal,
                         msg='Data frames not equal on index {0}, col {1}'.format(index, col))
        
    def test_PE_F(self):
        for adaptive_weights in [False, 'ridge', 'decisiontree', 'ridgeCV','decisiontreeCV']: #
            is_correct = 1
            try:
                model = None
                if adaptive_weights == False:
                    df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                    holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                          num_cov=7, min_val=0,
                                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                    covar_importance = np.array([4,3,2,1,0,0,0])
                    weight_array = covar_importance/covar_importance.sum()
                    model = matching.FLAME(repeats=False, verbose=0,adaptive_weights =adaptive_weights)
                    model.fit(holdout_data=holdout,weight_array = list(weight_array))
                    output = model.predict(df)
                else:
                    df, true_TE = generate_uniform_given_importance()
                    holdout, true_TE = generate_uniform_given_importance()
                    model = matching.FLAME(repeats=False, verbose=0,adaptive_weights =adaptive_weights)
                    model.fit(holdout_data=holdout)
                    output = model.predict(df)

                if check_statistics(model):
                    is_correct = 0
                    

            except (KeyError, ValueError):
                is_correct = 0

            self.assertEqual(1, is_correct,
                             msg='FLAME-Error when we use PE method: {0} '.format(str(adaptive_weights)))

    def test_datasets_F(self):
        df_path = os.path.join((os.path.dirname(__file__)), 'basicTestData.csv')

        for gen in [generate_uniform_given_importance,generate_binomial_given_importance,generate_binomial_decay_importance,df_path]:
            is_correct = 1
            try:
                df = None
                holdout = None
                if type(gen) != str:
                    df, true_TE = gen()
                    holdout, true_TE = gen()
                else:
                    df  = gen
                    holdout = gen
                model = matching.FLAME(repeats=False)
                model.fit(holdout_data=False)
                output = model.predict(df)
            
                if check_statistics(model):
                    is_correct = 0
                    break

            except (KeyError, ValueError):
                is_correct = 0

            self.assertEqual(1, is_correct,
                             msg='FLAME-Error when we use the dataset generated by {0} '.format(str(gen)))
    
    def test_repeats_F(self):
        #Test other parameters
        df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
        holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
        is_correct = 1
        try:
            model = matching.FLAME(repeats=True)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='FLAME-Error when repeat = True')
        
    def test_verbose_F(self):
        #Test verbose
        df, true_TE = generate_uniform_given_importance()
        for verbose in [0,1,2,3]:
            is_correct = 1
            try:
                df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000,
                                                              num_cov=7, min_val=0,
                                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                      num_cov=7, min_val=0,
                                                          max_val=3, covar_importance=[4,3,2,1,0,0,0])
                covar_importance = np.array([4,3,2,1,0,0,0])
                weight_array = covar_importance/covar_importance.sum()
                model = matching.FLAME(missing_data_replace = 2, want_bf = True, verbose = verbose)
                model.fit(holdout_data=holdout)
                output = model.predict(df)
                model = matching.FLAME(repeats=True,verbose=verbose)
                model.fit(holdout_data=0.5)
                output = model.predict(df)
                if check_statistics(model):
                    is_correct = 0
                    break
            except (KeyError, ValueError):
                is_correct = 0

            self.assertEqual(1, is_correct, msg='FLAME-Error when verbose ={0}'.format(verbose))
            
        
    def test_data_split_F(self):
        #Test data split
        df, true_TE = generate_uniform_given_importance(num_control=3000, num_treated=3000)

        is_correct = 1
        try:
            for holdout in [0.3,0.5,0.7]:
                model = matching.FLAME(repeats=True)
                model.fit(holdout_data=holdout)
                output = model.predict(df)
                if check_statistics(model):
                    is_correct = 0
                    break
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='FLAME-Error when holdout = {0}'.format(holdout))

        
    def test_miss_data_F(self):
        is_correct = 1
        try:
            for missing_holdout_replace in [0,1,2]:
                for missing_data_replace in [0,1,2,3]:
                    df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000)
                    #Create missing df
                    m,n = df.shape
                    for i in range(int(m/100)):
                        for j in [0,int(n/2)]:
                            df.iloc[i,j] = np.nan
                    holdout = df.copy()

                    model = matching.FLAME(missing_holdout_replace = missing_holdout_replace,missing_data_replace=missing_data_replace )
                    model.fit(holdout_data=holdout)
                    output = model.predict(df)
                    if check_statistics(model):
                        is_correct = 0
                        break

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='FLAME-Error when do missing data'\
                             'handling with missing_holdout_replace = {0},missing_data_replace{1}'.format(missing_holdout_replace,missing_data_replace))
        
    def test_want_pebf_F(self):
        #Test
        df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
        holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])

        is_correct = 1
        try:
            for want_pe in [False, True]:
                for want_bf in [False, True]:
                    model = matching.FLAME(want_pe=want_pe,want_bf=want_bf)
                    model.fit(holdout_data=holdout)
                    output = model.predict(df)
                    if check_statistics(model) or (want_pe and len(model.pe_each_iter)==0) or (want_bf and len(model.bf_each_iter)==0):
                        is_correct = 0
                        break

        except (KeyError, ValueError):
            is_wrong = 0
        self.assertEqual(1, is_correct, msg='FLAME Error when want_pe = {0} want_bf = {1}'.format(str(want_pe),str(want_bf)))
        
    def test_pre_dame_F(self):
        df, true_TE = generate_uniform_given_importance(num_control=500, num_treated=500,
                                  num_cov=7, min_val=0,
                                  max_val=3, covar_importance=[4,3,2,1,0,0,0])
        holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                                  max_val=3, covar_importance=[4,3,2,1,0,0,0])
        is_correct = 1
        try:
            covar_importance = np.array([4,3,2,1,0,0,0])
            weight_array = covar_importance/covar_importance.sum()
            for x in [False, True]:
                for y in [False, True]:
                    model1 = matching.FLAME(repeats=x,want_pe = y, want_bf = y,verbose=0,adaptive_weights = False)
                    model1.fit(holdout_data=holdout,weight_array = list(weight_array))
                    output = model1.predict(df, pre_dame = True)
                    model2 = matching.FLAME(repeats=x, want_pe = y, want_bf = y,verbose=0,adaptive_weights = 'decisiontreeCV')
                    model2.fit(holdout_data=holdout)
                    output = model2.predict(df, pre_dame = True)

                    output = model2.predict(df, pre_dame = True)
                    if check_statistics(model1) or check_statistics(model2) :
                        is_correct = 0
                
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct,
                         msg='FLAME-Error when we use pre_dame')
        
    def test_other_param_F(self):
        is_correct = 1
        try:
            df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000,
                                                  num_cov=7, min_val=0,
                                                  max_val=3, covar_importance=[4,3,2,1,0,0,0])
            holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                  num_cov=7, min_val=0,
                                                      max_val=3, covar_importance=[4,3,2,1,0,0,0])
            
            model = matching.FLAME( early_stop_pe= 1, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.FLAME( stop_unmatched_c= True, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.FLAME(stop_unmatched_t= True, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.FLAME(early_stop_un_c_frac = 0.5, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.FLAME(early_stop_un_t_frac = 0.5, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.FLAME(early_stop_iterations= 2, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
                
        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='FLAME-Error when other parameters')
    
    def test_no_matching_F(self):
        #Test data split
        df, true_TE = generate_uniform_given_importance(num_control=3000, num_treated=3000)

        is_correct = 1
        try:
            df = pd.DataFrame([[1,2,0,1.0],[3,4,1,2.0],[5,6,0,5.0],[7,8,1,8.0],[9,10,1,10.0]])
            df.columns = ['cov1','cov2','treated','outcome']
            holdout = df.copy()
            model = matching.FLAME(repeats=True)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='FLAME-Error when no matching')
        
    def test_miss_data_indicator_F(self):
        is_correct = 1
        try:
            df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000)
            #Create missing df
            m,n = df.shape
            for i in range(int(m/100)):
                for j in [0,int(n/2)]:
                    df.iloc[i,j] = 'a'
            holdout = df.copy()

            model = matching.FLAME(missing_indicator= 'a', missing_holdout_replace = 1,missing_data_replace=1 )
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='FLAME-Error when missing_indicator')
    def test_has_unmatched_units_F(self):
        #Test data split
        is_correct = 1
        try:
            df = pd.DataFrame([[1,2,0,1.0],[3,4,0,2.0],[5,6,0,5.0],[7,8,0,8.0],[9,10,1,10.0],[9,20,0,10.0]])
            df.columns = ['cov1','cov2','treated','outcome']
            holdout = df.copy()
            model = matching.FLAME(repeats=True)
            model.fit(holdout_data=holdout)
            output = model.predict(df)

            if check_statistics(model):
                is_correct = 0
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='FLAME-Error when no matching')

class TestDame(unittest.TestCase):
            
    def test_PE_F(self):
        for adaptive_weights in [ 'ridge', 'decisiontree', 'ridgeCV','decisiontreeCV']: #False,
            is_correct = 1
            try:
                model = None
                if adaptive_weights == False:
                    df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=7, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                    holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                          num_cov=7, min_val=0,
                                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                    covar_importance = np.array([4,3,2,1,0,0,0])
                    weight_array = covar_importance/covar_importance.sum()
                    model = matching.DAME(repeats=False, verbose=0,adaptive_weights =adaptive_weights)
                    model.fit(holdout_data=holdout,weight_array = list(weight_array))
                    output = model.predict(df)
                else:
                    df, true_TE = generate_uniform_given_importance()
                    holdout, true_TE = generate_uniform_given_importance()
                    model = matching.DAME(repeats=False, verbose=0,adaptive_weights =adaptive_weights)
                    model.fit(holdout_data=holdout)
                    output = model.predict(df)

                if check_statistics(model):
                    is_correct = 0
                    break

            except (KeyError, ValueError):
                is_correct = 0


            self.assertEqual(1, is_correct,
                             msg='DAME-Error when we use PE method: {0} '.format(adaptive_weights))

    def test_datasets_F(self):
        df_path = os.path.join((os.path.dirname(__file__)), 'basicTestData.csv')
        for gen in [generate_uniform_given_importance,generate_binomial_given_importance,generate_binomial_decay_importance,df_path]:
            is_correct = 1
            try:
                df = None
                holdout = None
                if type(gen) != str:
                    df, true_TE = gen()
                    holdout, true_TE = gen()
                else:
                    df  = gen
                    holdout = gen
                model = matching.DAME(repeats=False)
                model.fit(holdout_data=False)
                output = model.predict(df)
        
                if check_statistics(model):
                    is_correct = 0
                    break

            except (KeyError, ValueError):
                is_correct = 0

            self.assertEqual(1, is_correct,
                             msg='DAME-Error when we use the dataset generated by {0} '.format(str(gen)))
    
    def test_repeats_F(self):
        #Test other parameters
        df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                               num_cov=7, min_val=0,
                                               max_val=3, covar_importance=[4,3,2,1,0,0,0])
        holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                               num_cov=7, min_val=0,
                                               max_val=3, covar_importance=[4,3,2,1,0,0,0])
        is_correct = 1
        try:
            model = matching.DAME(repeats=True)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='DAME-Error when repeat = True')

    def test_verbose_F(self):
        #Test verbose
        df, true_TE = generate_uniform_given_importance()
        for verbose in [0,1,2,3]:
            is_correct = 1
            try:
                df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000,
                                                              num_cov=7, min_val=0,
                                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                      num_cov=7, min_val=0,
                                                          max_val=3, covar_importance=[4,3,2,1,0,0,0])
                covar_importance = np.array([4,3,2,1,0,0,0])
                weight_array = covar_importance/covar_importance.sum()
                model = matching.DAME(missing_data_replace = 2, want_bf = True, verbose = verbose)
                model.fit(holdout_data=holdout)
                output = model.predict(df)

                model = matching.DAME(verbose=verbose) # repeats = True
                model.fit(holdout_data=0.5)
                output = model.predict(df)
                if check_statistics(model):
                    is_correct = 0
                    break
            except (KeyError, ValueError):
                is_correct = 0

            self.assertEqual(1, is_correct, msg='DAME-Error when verbose ={0}'.format(verbose))
            
        
    def test_data_split_F(self):
        #Test data split
        df, true_TE = generate_uniform_given_importance(num_control=3000, num_treated=3000)

        is_correct = 1
        try:
            for holdout in [0.3,0.5,0.7]:
                model = matching.DAME(repeats=True)
                model.fit(holdout_data=holdout)
                output = model.predict(df)
                if check_statistics(model):
                    is_correct = 0
                    break
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='DAME-Error when holdout = {0}'.format(holdout))

        
    def test_miss_data_F(self):

        is_correct = 1
        try:
            for missing_holdout_replace in [0,1,2]:
                for missing_data_replace in [0,1,2,3]:
                    #Test missig data handling
                    df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000)
                    #Create missing df
                    m,n = df.shape
                    for i in range(int(m/10)):
                        for j in [0,int(n/2)]:
                            df.iloc[i,j] = np.nan
                    holdout = df.copy()
                    model = matching.DAME(repeats = False,missing_holdout_replace = missing_holdout_replace,missing_data_replace=missing_data_replace )
                    model.fit(holdout_data=holdout)
                    output = model.predict(df)
                    if check_statistics(model):
                        is_correct = 0
                        break

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='DAME-Error when do missing data'\
                             'handling with missing_holdout_replace = {0},missing_data_replace{1}'.format(missing_holdout_replace,missing_data_replace))
        
    def test_want_pebf_F(self):
        #Test
        df, true_TE = generate_uniform_given_importance(num_control=3000, num_treated=3000,
                                              num_cov=6, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0])
        holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                              num_cov=6, min_val=0,
                                              max_val=3, covar_importance=[4,3,2,1,0,0])

        is_correct = 1
        try:
            for want_pe in [False, True]:
                for want_bf in [False, True]:
                    model = matching.DAME(want_pe=want_pe,want_bf=want_bf)
                    model.fit(holdout_data=holdout)
                    output = model.predict(df)
                    if check_statistics(model) or (want_pe and len(model.pe_each_iter)==0) or (want_bf and len(model.bf_each_iter)==0):
                        is_correct = 0
                        break

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='DAME Error when want_pe = {0} want_bf = {1}'.format(str(want_pe),str(want_bf)))
        
    def test_other_param_F(self):
        is_correct = 1
        try:
            df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000,
                                                  num_cov=7, min_val=0,
                                                  max_val=3, covar_importance=[4,3,2,1,0,0,0])
            holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                  num_cov=7, min_val=0,
                                                      max_val=3, covar_importance=[4,3,2,1,0,0,0])

            model = matching.DAME( early_stop_pe= 1, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.DAME( stop_unmatched_c= True, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.DAME(stop_unmatched_t= True, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.DAME(early_stop_un_c_frac = 0.5, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.DAME(early_stop_un_t_frac = 0.5, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0
            model = matching.DAME(early_stop_iterations= 2, verbose=0)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='DAME-Error when other parameters')
        
    def test_no_matching_F(self):
        #Test data split
        df, true_TE = generate_uniform_given_importance(num_control=3000, num_treated=3000)

        is_correct = 1
        try:
            df = pd.DataFrame([[1,2,0,1.0],[3,4,1,2.0],[5,6,0,5.0],[7,8,1,8.0],[9,10,1,10.0]])
            df.columns = ['cov1','cov2','treated','outcome']
            holdout = df.copy()
            model = matching.DAME(repeats=True)
            model.fit(holdout_data=holdout)
            output = model.predict(df)
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='DAME-Error when no matching')

    def test_miss_data_indicator_F(self):
        is_correct = 1
        try:
            df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000)
            #Create missing df
            m,n = df.shape
            for i in range(int(m/100)):
                for j in [0,int(n/2)]:
                    df.iloc[i,j] = 'a'
            holdout = df.copy()

            model = matching.DAME(missing_indicator= 'a', missing_holdout_replace = 1,missing_data_replace=1 )
            model.fit(holdout_data=holdout)
            output = model.predict(df)
            if check_statistics(model):
                is_correct = 0

        except (KeyError, ValueError):
            is_correct = 0
        self.assertEqual(1, is_correct, msg='DAME-Error when missing_indicator')

    def test_has_unmatched_units_F(self):
        #Test data split
        is_correct = 1
        try:
            df = pd.DataFrame([[1,2,0,1.0],[3,4,0,2.0],[5,6,0,5.0],[7,8,0,8.0],[9,10,1,10.0],[9,20,0,10.0]])
            df.columns = ['cov1','cov2','treated','outcome']
            holdout = df.copy()
            model = matching.DAME(repeats=True)
            model.fit(holdout_data=holdout)
            output = model.predict(df)

            if check_statistics(model):
                is_correct = 0
        except (KeyError, ValueError):
            is_correct = 0

        self.assertEqual(1, is_correct, msg='DAME-Error when no matching')



class Test_exceptions(unittest.TestCase):
    
    def test_false_dataset(self):
        def broken_false_dataset():
            df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000)
            holdout, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000)
            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(False)

            
        with self.assertRaises(Exception) as false_dataset:
            broken_false_dataset()
            
        self.assertTrue("Need to specify either csv file name or pandas data "\
                        "frame in parameter 'input_data'" in str(false_dataset.exception))
        
    def test_false_early_stop_un_t_frac(self):
        def broken_early_stop_un_t_frac():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(early_stop_un_t_frac = -1)
            model.fit(holdout_data=df)
            output = model.predict(df)

        with self.assertRaises(Exception) as early_stop_un_t_frac:
            broken_early_stop_un_t_frac()
            
        self.assertTrue('The value provided for the early stopping critera '\
                        'of proportion of unmatched treatment units needs to '\
                        'be between 0.0 and 1.0' in str(early_stop_un_t_frac.exception))
    
    def test_false_early_stop_un_c_frac(self):
        def broken_early_stop_un_c_frac():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(early_stop_un_c_frac = -1)
            model.fit(holdout_data=df)
            output = model.predict(df)

        with self.assertRaises(Exception) as early_stop_un_c_frac:
            broken_early_stop_un_c_frac()
            
        self.assertTrue('The value provided for the early stopping critera '\
                        'of proportion of unmatched control units needs to '\
                        'be between 0.0 and 1.0' in str(early_stop_un_c_frac.exception))
        
        
    def test_false_early_stop_iterations(self):
        def broken_early_stop_iterations():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(early_stop_iterations = True)
            model.fit(holdout_data=df)
            output = model.predict(df)

        with self.assertRaises(Exception) as early_stop_iterations:
            broken_early_stop_iterations()
            
        self.assertTrue('The value provided for early_stop_iteration needs '\
                        'to be an integer number of iterations, or False if '\
                        'not stopping early based on the number of iterations' in str(early_stop_iterations.exception))
    def test_false_early_stop_pe_frac(self):
        def broken_early_stop_pe_frac():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(early_stop_pe_frac = 123)
            model.fit(holdout_data=df)
            output = model.predict(df)

        with self.assertRaises(Exception) as early_stop_pe_frac:
            broken_early_stop_pe_frac()
            
        self.assertTrue('The value provided for the early stopping critera of'\
                        ' PE needs to be between 0.0 and 1.0' in str(early_stop_pe_frac.exception))


    def test_false_weight_array_type(self):
        def broken_weight_array_type():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(adaptive_weights = False)
            model.fit(holdout_data=df, weight_array = np.array([1,2,3,4,5]))
            output = model.predict(df)

        with self.assertRaises(Exception) as weight_array_type:
            broken_weight_array_type()
            
        self.assertTrue('Invalid input error. A weight array of type'\
                            'array needs to be provided when the'\
                            'parameter adaptive_weights == True' in str(weight_array_type.exception))

    def test_false_weight_array_len(self):
        def broken_weight_array_len():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(adaptive_weights = False)
            model.fit(holdout_data=df, weight_array = [1])
            output = model.predict(df)

        with self.assertRaises(Exception) as weight_array_len:
            broken_weight_array_len()
            
        self.assertTrue('Invalid input error. Weight array size not equal'\
                            ' to number of columns in dataframe' in str(weight_array_len.exception))
        
        
    def test_false_weight_array_sum(self):
        def broken_weight_array_sum():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(adaptive_weights = False)
            model.fit(holdout_data=df, weight_array = [1,1,1,1])
            output = model.predict(df)

        with self.assertRaises(Exception) as weight_array_sum:
            broken_weight_array_sum()
            
        self.assertTrue('Invalid input error. Weight array values must '\
                            'sum to 1.0' in str(weight_array_sum.exception))
        
        
        
    def test_false_alpha(self):
        def broken_alpha():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(adaptive_weights = 'ridge',alpha = -10)
            model.fit(holdout_data=df)
            output = model.predict(df)

        with self.assertRaises(Exception) as alpha:
            broken_alpha()
            
        self.assertTrue('Invalid input error. The alpha needs to be '\
                            'positive for ridge regressions.' in str(alpha.exception))
        
    def test_false_adaptive_weights(self):
        def broken_adaptive_weights():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            model = matching.FLAME(adaptive_weights = 'safdsaf')
            model.fit(holdout_data=df)
            output = model.predict(df)

        with self.assertRaises(Exception) as adaptive_weights:
            broken_adaptive_weights()
            
        self.assertTrue("Invalid input error. The acceptable values for "\
                            "the adaptive_weights parameter are 'ridge', "\
                            "'decisiontree', 'decisiontreeCV', or 'ridgeCV'. Additionally, "\
                            "adaptive-weights may be 'False' along "\
                            "with a weight array" in str(adaptive_weights.exception))
        
    def test_false_data_len(self):
        def broken_data_len():
            df, true_TE = generate_uniform_given_importance(num_control=1000, num_treated=1000,
                                                  num_cov=7, min_val=0,
                                                  max_val=3, covar_importance=[4,3,2,1,0,0,0])
            holdout, true_TE = generate_uniform_given_importance()
            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(df)

        with self.assertRaises(Exception) as data_len:
            broken_data_len()
            
        self.assertTrue('Invalid input error. The holdout and main '\
                            'dataset must have the same number of columns' in str(data_len.exception))
    
    def test_false_column_match(self):
        def broken_column_match():
            df, true_TE = generate_uniform_given_importance()
            holdout, true_TE = generate_uniform_given_importance()
            set_ = holdout.columns
            set_ = list(set_)
            set_[0] = 'dasfadf'
            holdout.columns  = set_
            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(df)

        with self.assertRaises(Exception) as column_match:
            broken_column_match()
            
        self.assertTrue('Invalid input error. The holdout and main '\
                            'dataset must have the same columns' in str(column_match.exception))
    def test_false_C(self):
        def broken_C():
            df, true_TE = generate_uniform_given_importance()
            holdout, true_TE = generate_uniform_given_importance()

            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(df,C = -1)

        with self.assertRaises(Exception) as C:
            broken_C()
            
        self.assertTrue('The C, or the hyperparameter to trade-off between'\
                           ' balancing factor and predictive error must be '\
                           ' nonnegative. 'in str(C.exception))
    def test_false_missing_data_replace(self):
        def broken_missing_data_replace():
                df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                              num_cov=7, min_val=0,
                                                              max_val=3, covar_importance=[4,3,2,1,0,0,0])
                holdout, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100,
                                                      num_cov=7, min_val=0,
                                                          max_val=3, covar_importance=[4,3,2,1,0,0,0])
                covar_importance = np.array([4,3,2,1,0,0,0])
                weight_array = covar_importance/covar_importance.sum()
                model = matching.FLAME(missing_data_replace = 2, adaptive_weights =False)
                model.fit(holdout_data=holdout,weight_array = list(weight_array))
                output = model.predict(df)

        with self.assertRaises(Exception) as missing_data_replace:
            broken_missing_data_replace()
            
        self.assertTrue('Invalid input error. We do not support missing data '\
                        'handing in the fixed weights version of algorithms'in str(missing_data_replace.exception))
        
    def test_false_treatment_column_name(self):
        def broken_treatment_column_name():
            df, true_TE = generate_uniform_given_importance()
            holdout, true_TE = generate_uniform_given_importance()
            model = matching.FLAME()
            model.fit(holdout_data=holdout,treatment_column_name =  "sadfdag")
            output = model.predict(df)

        with self.assertRaises(Exception) as treatment_column_name:
            broken_treatment_column_name()
            
        self.assertTrue('Invalid input error. Treatment column name does not'\
                        ' exist' in str(treatment_column_name.exception))

    def test_false_outcome_column_name(self):
        def broken_outcome_column_name():
            df, true_TE = generate_uniform_given_importance()
            holdout, true_TE = generate_uniform_given_importance()
            model = matching.FLAME()
            model.fit(holdout_data=holdout,outcome_column_name =  "sadfdag")
            output = model.predict(df)

        with self.assertRaises(Exception) as outcome_column_name:
            broken_outcome_column_name()
            
        self.assertTrue('Invalid input error. Outcome column name does not'\
                        ' exist' in str(outcome_column_name.exception))
        
    def test_false_treatment_column_name_value(self):
        def broken_treatment_column_name_value():
            df, true_TE = generate_uniform_given_importance()
            holdout, true_TE = generate_uniform_given_importance()
            df.loc[0,'treated'] = 4
            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(df)

        with self.assertRaises(Exception) as treatment_column_name_value:
            broken_treatment_column_name_value()
        self.assertTrue('Invalid input error. All rows in the treatment '\
                        'column must have either a 0 or a 1 value.' in str(treatment_column_name_value.exception))
        

        
    def test_false_data_type(self):
        def broken_data_type():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            holdout = df.copy()
            df.iloc[0,0] = 's'
            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(df)

        with self.assertRaises(Exception) as _data_type:
            broken_data_type()

        self.assertTrue('Invalid input error on matching dataset. Ensure all inputs asides from '\
                        'the outcome column are integers, and if missing' \
                        ' values exist, ensure they are handled.' in str(_data_type.exception))
    def test_false_holdout_type(self):
        def broken_holdout_type():
            df, true_TE = generate_uniform_given_importance(num_control=100, num_treated=100)
            holdout = df.copy()
            holdout.iloc[0,0] = 's'
            model = matching.FLAME()
            model.fit(holdout_data=holdout)
            output = model.predict(df)

        with self.assertRaises(Exception) as holdout_type:
            broken_holdout_type()

        self.assertTrue('Invalid input error on holdout dataset. Ensure all inputs asides from '\
                                'the outcome column are integers, and if missing' \
                                ' values exist, ensure they are handled.' in str(holdout_type.exception))

    def test_false_ATE_input(self):
        def broken_ATE_input():
            ATE(1)

        with self.assertRaises(Exception) as ATE_input:
            broken_ATE_input()
        self.assertTrue("The matching_object input parameter needs to be "\
                            "of type DAME or FLAME" in str(ATE_input.exception))

    def test_false_ATE_input_model(self):
        def broken_ATE_input_model():
            model = matching.FLAME()
            ATE(model)
        with self.assertRaises(Exception) as ATE_input_model:
            broken_ATE_input_model()
        self.assertTrue("This function can be only called after a match has "\
                           "been formed using the .fit() and .predict() functions" in str(ATE_input_model.exception))



# -*- coding: utf-8 -*-
"""Helper functions for the flame_algorithm.py and dame_algorithm.py files"""

# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

import numpy as np
import pandas as pd

from sklearn.linear_model import Ridge
from sklearn.model_selection import cross_val_score

from sklearn.tree import DecisionTreeRegressor
from sklearn.metrics import mean_squared_error

from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer

def verbose_output(iteration_number, num_matched_groups, num_unmatched_t,
                   num_unmatched, orig_len_df_all, tot_treated, pe,
                   prev_iter_num_unmatched, curr_covar_set):
    '''
    Prints progress of matching algorithm along various metrics
    '''
    print("Iteration number: ", iteration_number)
    print("\tNumber of matched groups formed in total: ", num_matched_groups)
    print("\tUnmatched treated units: ", num_unmatched_t, "out of a total of ",
          tot_treated, "treated units")
    print("\tUnmatched control units: ", num_unmatched-num_unmatched_t,
          "out of a total of ", orig_len_df_all-tot_treated, "control units")
    print("\tPredictive error of covariates chosen this iteration: ", pe)
    print("\tNumber of matches made in this iteration: ",
          prev_iter_num_unmatched - num_unmatched)
    print("\tNumber of matches made so far: ",
          orig_len_df_all - num_unmatched)
    print("\tIn this iteration, the covariates dropped are: ", curr_covar_set)


def compute_bf(matched_rows, treatment_column_name, df_unmatched):
    '''
    Helper function to compute the balancing factor
    '''

    mg_treated = matched_rows[treatment_column_name].sum()
    mg_control = len(matched_rows) - mg_treated
    available_treated = df_unmatched[treatment_column_name].sum()
    available_control = len(df_unmatched) - available_treated

    if (available_treated != 0 and available_control != 0):
        return (mg_treated/available_treated + mg_control/available_control)
    elif available_treated == 0:
        return mg_control/available_control
    return mg_treated/available_treated


def find_pe_for_covar_set(df_holdout, treatment_column_name,
                          outcome_column_name, s, adaptive_weights,
                          alpha_given):
    '''
    this is a helper function to decide_drop that will find pe of a given s
    '''
    # The iteration and mean of array is only used when doing MICE on holdout
    pe_array = []
    for i in range(len(df_holdout)):


        x_treated, x_control, y_treated, y_control = separate_dfs(
            df_holdout[i], treatment_column_name, outcome_column_name, s)

        # error check. If this is true, we stop matching.
        if type(x_treated) == bool:
            return False

        # binarize holdout dataset if categorical:
        if (adaptive_weights == "decisiontree" or adaptive_weights == "decisiontreeCV"):

            # first binarize non-binary columns in the treated dataset
            bool_cols = bool_cols = [col for col in x_treated
                                     if np.isin(x_treated[col].unique(), [0, 1]).all()]
            non_bool_cols = x_treated.columns.difference(bool_cols)
            binarized_df = pd.get_dummies(x_treated.loc[:, non_bool_cols].astype(str))
            x_treated = pd.concat([binarized_df, x_treated.loc[:, bool_cols]], axis=1)

            # binarize non-binary columns in the control dataset
            bool_cols = bool_cols = [col for col in x_control
                                     if np.isin(x_control[col].unique(), [0, 1]).all()]
            non_bool_cols = x_control.columns.difference(bool_cols)
            binarized_df = pd.get_dummies(x_control.loc[:, non_bool_cols].astype(str))
            x_control = pd.concat([binarized_df, x_control.loc[:, bool_cols]], axis=1)

        if adaptive_weights in ["ridge", "ridgeCV"]:
            clf = Ridge(alpha=alpha_given)
        elif adaptive_weights in ["decisiontree", "decisiontreeCV"]:
            clf = DecisionTreeRegressor()
        else:
            return False

        if adaptive_weights in ["ridge", "decisiontree"]:

            # Calculate treated MSE
            clf.fit(x_treated, y_treated)
            predicted = clf.predict(x_treated)
            mse_treated = mean_squared_error(y_treated, predicted)

            # Calculate control MSE
            clf.fit(x_control, y_control)
            predicted = clf.predict(x_control)
            mse_control = mean_squared_error(y_control, predicted)

        else:
            # calculate MSE via cross validation.
            mse_treated = -1*np.mean(cross_val_score(clf, x_treated, y_treated,
                                                     scoring='neg_mean_squared_error',
                                                     cv=5))
            mse_control = -1*np.mean(cross_val_score(clf, x_control, y_control,
                                                     scoring='neg_mean_squared_error',
                                                     cv=5))

        pe_array.append(mse_treated + mse_control)

    return np.mean(pe_array)

def create_mice_dfs(df_holdout, num_imputes, outcome_col_name):
    '''
    This creates num_imputes number of imputed datasets
    '''
    df_holdout_array = []
    for i in range(num_imputes):
        imp = IterativeImputer(max_iter=10, random_state=i,
                               estimator=DecisionTreeRegressor())
        imp.fit(df_holdout)
        tmp_df = pd.DataFrame(data=np.round(imp.transform(df_holdout)),
                              columns=df_holdout.columns,
                              index=df_holdout.index)
        # convert floats to ints because MICE creates floats
        cols = list(tmp_df.columns)
        cols.remove(outcome_col_name)
        tmp_df[cols] = tmp_df[cols].astype('int64')
        df_holdout_array.append(tmp_df)

    return df_holdout_array

def separate_dfs(df_holdout, treatment_col_name, outcome_col_name,
                 covs_include):
    """
    This function serves to create the control/treatment dfs for use
    in the decide_drop functions in flame and in dame.
    """
    #X-treated is the df that has rows where treated col = 1 and
    # all cols except: outcome/treated/the covs being dropped
    x_treated = df_holdout.loc[df_holdout[treatment_col_name] == 1,
                               df_holdout.columns.difference([outcome_col_name,
                                                              treatment_col_name] + list(covs_include))]

    #X-control is the df that has rows where treated col = 0 and
    # all cols except: outcome/treated/the covs being dropped
    x_control = df_holdout.loc[df_holdout[treatment_col_name] == 0,
                               df_holdout.columns.difference([outcome_col_name,
                                                              treatment_col_name] + list(covs_include))]

    y_treated = df_holdout.loc[df_holdout[treatment_col_name] == 1,
                               outcome_col_name]

    y_control = df_holdout.loc[df_holdout[treatment_col_name] == 0,
                               outcome_col_name]

    # error check. If this is true, we stop matching.
    if (len(x_treated) == 0 or len(x_control) == 0 or \
        len(y_treated) == 0 or len(y_control) == 0 or \
        len(x_treated.columns) == 0 or len(x_control.columns) == 0):
        return False, False, False, False

    return x_treated, x_control, y_treated, y_control

def stop_iterating(early_stops, df_unmatched, repeats, treat_col_name,
                   orig_len_df_all, h, orig_tot_treated, consider_dropping):
    """
    This function is called during the iterations of DAME and FLAME to see
    if any stopping criteria have been met
    """

    # Iterates while there are units to match to match in
    try:
        if ((early_stops.unmatched_t == True or repeats == False) and
            (1 not in df_unmatched[treat_col_name].values)):
            print(orig_len_df_all - len(df_unmatched), "units matched. "\
                  "We finished with no more treated units to match")
            return True

        if ((early_stops.unmatched_c == True or repeats == False) and
            (0 not in df_unmatched[treat_col_name].values)):
            print(orig_len_df_all - len(df_unmatched), "units matched. "\
                  "We finished with no more control units to match")
            return True

    except TypeError:
        return True
        
    # Hard stop criteria: stop when there are no more units to match
    if (len(df_unmatched) == 0):
        print("All units have been matched.")
        return True

    # Hard stop criteria: exceeded the number of iters user asked for?
    if (early_stops.iterations != False and early_stops.iterations == h):
        print((orig_len_df_all - len(df_unmatched)), "units matched. "\
              "We stopped before doing iteration number: ", h)
        return True

    # Hard stop criteria: met the threshold of unmatched items to stop?
    if (early_stops.un_t_frac != False or early_stops.un_c_frac != False):
        unmatched_treated = df_unmatched[treat_col_name].sum()
        unmatched_control = len(df_unmatched) - unmatched_treated
        orig_tot_control = orig_len_df_all - orig_tot_treated
        if (early_stops.un_t_frac != False and \
            unmatched_treated/orig_tot_treated < early_stops.un_t_frac):
            print("We stopped the algorithm when ",
                  unmatched_treated/orig_tot_treated, "of the treated "\
                  "units remained unmatched")
            return True

        elif (early_stops.un_c_frac != False and \
            unmatched_control/orig_tot_control < early_stops.un_c_frac):
            print("We stopped the algorithm when ",
                  unmatched_control/orig_tot_control, "of the control "\
                  "units remained unmatched")
            return True

    # quit if there are no more covariate sets to choose from
    if (len(consider_dropping) == 1):
        print((orig_len_df_all - len(df_unmatched)), "units matched. "\
              "No more covariate sets to consider dropping")
        return True

    return False
        


import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()



setuptools.setup(name='dame_flame',
      version='0.41',
      description='Causal Inference Covariate Matching',
      long_description=long_description,
      keywords='Causal Inference Matching Econometrics Data Machine Learning FLAME DAME',
      url='https://github.com/almost-matching-exactly/DAME-FLAME-Python-Package',
      author='Neha Gupta',
      author_email='neha.r.gupta@duke.edu',
      license='MIT',
      packages=setuptools.find_packages(),
      install_requires=[
          'scikit-learn>=0.23.2',
          'pandas>=0.16.5',
          'numpy>=1.6.1'
      ],
     long_description_content_type="text/markdown",
     classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    )



# -*- coding: utf-8 -*-
"""
DAME and FLAME Matching Methods for Causal Inference. Starts matching methods.
"""
# author: Neha Gupta, Duke University
# Copyright Duke University 2020
# License: MIT

import numpy as np
from . import data_cleaning
from . import dame_algorithm
from . import flame_algorithm
from . import flame_dame_helpers

class MatchParent:
    """ Matching via DAME or FLAME algorithms

    MatchParent aims to create and object for matching a given dataset.
    Read more in the documentation User Guide:
    https://almost-matching-exactly.github.io/DAME-FLAME-Python-Package/

    Parameters
    -----------
    adaptive_weights (bool, str): Weight dropping method. False, 'ridge',
        'decisiontree', 'ridgeCV', or 'decisiontreeCV'.
    alpha (float): This is the alpha for ridge regression. We use the
        scikit package for ridge regression, so it is "regularization
        strength". Larger values specify stronger regularization.
        Must be positive float.
    repeats (bool): whether values for whom a MMG has been found can
        be used again and placed in an auxiliary matched group.
    verbose (default: 2): If 1, provides iteration num, if 2 provides
        iteration number and number of units left to match on every 10th
        iter, if 3 does this print on every iteration. If 0, nothing.
    early_stop_iterations (optional int): If provided, a number of iters
        to hard stop the algorithm after.
    stop_unmatched_c, stop_unmatched_t (bools): specifies whether
        the algorithm stops when there are no units remaining to match
    early_stop_un_c_frac, early_stop_un_t_frac (optional float,
        from 0.0 - 1.0): If provided, a fraction of unmatched control/
        treatment units. When threshold met, hard stop the algo.
    early_stop_pe: Whether the covariate set chosen to match
        on has a pe lower than the parameter early_stop_pe_frac, at
        which point the algorithm will stop.
    early_stop_pe_frac: If early_stop_pe is true,
        then if the covariate set chosen to match on has a PE lower than
        this value, the algorithm will stop
    missing_holdout_replace (0,1,2): default 0.
        if 0, assume no missing holdout data and proceed
        if 1, drop all missing_indicator values from holdout dataset
        if 2, do mice on holdout dataset for missing_holdout_imputations
        number of imputations
    missing_data_replace (0,1,2,3): default 0.
        if 0, assume no missing data in matching data and proceed
        if 1, drop all missing_indicator values from matching data
        if 2, replace all missing_indicator values with unique large vals
        so they essentially get skipped in the matching
        if 3, do mice on matching dataset for missing_data_imputations
        number of imputations.
    missing_holdout_imputations: If missing_holdout_replace=2, the number
        of imputations on the holdout set.
    missing_data_imputations: If missing_data_replace=3, the number of
        imputations on the matching set.
    missing_indicator: This is the character/number/np.nan that indicates
        missing vals in the holdout/matching data.
    want_pe: whether the output of the algorithm will include the
        predictive error of the covariate sets matched on in each iteration
    want_bf: whether the output will include the balancing factor of each
        iteration.
    """
    def __init__(self, adaptive_weights='ridge', alpha=0.1, repeats=True,
                 verbose=2, early_stop_iterations=False,
                 stop_unmatched_c=False, early_stop_un_c_frac=False,
                 stop_unmatched_t=False, early_stop_un_t_frac=False,
                 early_stop_pe=False, early_stop_pe_frac=0.01,
                 missing_indicator=np.nan, missing_data_replace=0,
                 missing_holdout_replace=0, missing_holdout_imputations=10,
                 missing_data_imputations=1, want_pe=False, want_bf=False):

        self.adaptive_weights = adaptive_weights
        self.alpha = alpha
        self.repeats = repeats
        self.verbose = verbose
        self.missing_indicator = missing_indicator
        self.missing_data_replace = missing_data_replace
        self.missing_holdout_replace = missing_holdout_replace
        self.missing_holdout_imputations = missing_holdout_imputations
        self.missing_data_imputations = missing_data_imputations
        self.early_stop_iterations = early_stop_iterations
        self.stop_unmatched_c = stop_unmatched_c
        self.early_stop_un_c_frac = early_stop_un_c_frac
        self.stop_unmatched_t = stop_unmatched_t
        self.early_stop_un_t_frac = early_stop_un_t_frac
        self.early_stop_pe = early_stop_pe
        self.early_stop_pe_frac = early_stop_pe_frac
        self.want_pe = want_pe
        self.want_bf = want_bf

    def fit(self, holdout_data=False, treatment_column_name='treated',
            outcome_column_name='outcome', weight_array=False):
        """
        Save model data that will be used to fit the matching method.

        Parameters
        ----------
        holdout_data : {string, dataframe, float, False }, default = False
            This is the holdout dataset. If a string is given, that should be
            the location of a CSV file to input. If a float between 0.0 and 1.0
            is given, that corresponds the percent of the input dataset to
            randomly select for holdout data. If False, the holdout data is
            equal to the entire input data.
        treatment_column_name : string, default="treated"
            the name of the column with a binary indicator for whether a row is
            a treatment or control unit.
        outcome_column_name : string, default="outcome"
            This is the name of the column with the outcome variable of each
            unit.
        weight_array : array, optional
            If adaptive_weights = False, these are the weights to the
            covariates in input_data, for the non-adaptive version of DAME.
            Must sum to 1. In this case, we do not use machine learning for the
            weights, they are manually entered as weight_array.

        """
        self.holdout_data = holdout_data
        self.treatment_column_name = treatment_column_name
        self.outcome_column_name = outcome_column_name
        self.weight_array = weight_array

class DAME(MatchParent):
    '''
    This is the class used for the DAME algorithm
    '''
    def predict(self, input_data):
        """
        Performs match and returns matched data.

        Parameters
        ----------
        input_data: {string, dataframe}, required parameter
            The dataframe on which to perform the matching, or the location of
            the CSV with the dataframe
        """
        self.input_data, self.holdout_data = data_cleaning.read_files(
            input_data, self.holdout_data)

        return_array = _DAME(self.input_data.copy(deep=True),
            self.holdout_data.copy(deep=True),
            self.treatment_column_name, self.weight_array,
            self.outcome_column_name, self.adaptive_weights, self.alpha,
            self.repeats, self.verbose, self.want_pe,
            self.early_stop_iterations, self.stop_unmatched_c,
            self.early_stop_un_c_frac, self.stop_unmatched_t,
            self.early_stop_un_t_frac, self.early_stop_pe,
            self.early_stop_pe_frac, self.want_bf,
            self.missing_indicator,
            self.missing_data_replace, self.missing_holdout_replace,
            self.missing_holdout_imputations, self.missing_data_imputations)

        self.bf_each_iter = None
        self.pe_each_iter = None

        # in the non-mice case:
        if (self.missing_data_replace != 3):
            self.df_units_and_covars_matched = return_array[0]
            self.groups_per_unit = self.df_units_and_covars_matched['weights']
            self.df_units_and_covars_matched = self.df_units_and_covars_matched.drop(columns = ['weights'])
            self.units_per_group = return_array[1]
            if (self.want_pe == True):
                self.pe_each_iter = return_array[2]
            if (self.want_bf == True):
                self.bf_each_iter = return_array[-1]
        else:
            # in the mice case:
            array_of_dfs = []
            array_of_groups_per_unit = []
            array_of_units_per_group = []
            for arr in return_array:
                temp_df = arr[0]
                array_of_groups_per_unit.append(temp_df['weights'])
                array_of_dfs.append(temp_df.drop(columns=['weights']))
                array_of_units_per_group.append(arr[1])
            self.groups_per_unit = array_of_groups_per_unit
            self.df_units_and_covars_matched = array_of_dfs
            self.units_per_group = array_of_units_per_group

        return self.df_units_and_covars_matched

class FLAME(MatchParent):
    '''
    The class used for the FLAME algorithm
    '''
    def predict(self, input_data, pre_dame=False, C=0.1):
        """
        Performs match and returns matched data.

        Parameters
        ----------
        input_data: {string, dataframe}, required parameter
            The dataframe on which to perform the matching, or the location of
            the CSV with the dataframe
        pre_dame (int, False): Indicates whether to switch to dame and after
            int number of iterations.
        C (float, 0.1): The tradeoff between PE and BF in computing MQ
        """
        self.input_data, self.holdout_data = data_cleaning.read_files(
            input_data, self.holdout_data)

        return_array = _FLAME(self.input_data.copy(deep=True),
            self.holdout_data.copy(deep=True), self.treatment_column_name,
            self.weight_array,self.outcome_column_name, self.adaptive_weights,
            self.alpha, self.repeats, self.verbose, self.want_pe,
            self.early_stop_iterations, self.stop_unmatched_c,
            self.early_stop_un_c_frac, self.stop_unmatched_t,
            self.early_stop_un_t_frac, self.early_stop_pe,
            self.early_stop_pe_frac, self.want_bf, self.missing_indicator,
            self.missing_data_replace, self.missing_holdout_replace,
            self.missing_holdout_imputations, self.missing_data_imputations,
            pre_dame, C)

        self.bf_each_iter = None
        self.pe_each_iter = None
		
        # in the non-mice case:
        if self.missing_data_replace != 3 and not pre_dame:
            self.df_units_and_covars_matched = return_array[0]
            self.groups_per_unit = self.df_units_and_covars_matched['weights']
            self.df_units_and_covars_matched = self.df_units_and_covars_matched.drop(columns=['weights'])
            self.units_per_group = return_array[1]
            if self.want_pe:
                self.pe_each_iter = return_array[2]
            if self.want_bf:
                self.bf_each_iter = return_array[-1]

        if (self.missing_data_replace != 3 and pre_dame==False):
            self.df_units_and_covars_matched = return_array[0]
            self.groups_per_unit = self.df_units_and_covars_matched['weights']
            self.df_units_and_covars_matched = self.df_units_and_covars_matched.drop(columns = ['weights'])
            self.units_per_group = return_array[1]
            if (self.want_pe == True):
                self.pe_each_iter = return_array[2]
            if (self.want_bf == True):
                self.bf_each_iter = return_array[-1]

        elif pre_dame:

            # the first few items all look the same, then the last item is from dame
            self.df_units_and_covars_matched = return_array[0]
            self.df_units_and_covars_matched = self.df_units_and_covars_matched.append(return_array[-1][0], sort=True)
            if self.repeats == True:
                # we have to aggregate the indexes appearing more than once,
                # those are the ones which were matched units in both dame and flame:

                df_wrepeats = self.df_units_and_covars_matched
                # this is a df without repeat indexes
                df_no_repeats = df_wrepeats[~df_wrepeats.index.duplicated(keep='first')].copy()

                #iterate through repeat indexes and sum weight column
                matched_twice_ind = df_wrepeats[df_wrepeats.index.duplicated()].index.unique()
                for index in matched_twice_ind:
                    df_no_repeats.loc[index, 'weights'] = df_wrepeats.loc[index].copy()['weights'].sum()

                self.df_units_and_covars_matched = df_no_repeats

            self.df_units_and_covars_matched.replace(np.nan, "*")
            self.groups_per_unit = self.df_units_and_covars_matched['weights']

            self.df_units_and_covars_matched = self.df_units_and_covars_matched.drop(columns = ['weights'])
            self.units_per_group = return_array[1]
            self.units_per_group += return_array[-1][1]
            if (self.want_pe == True):
                self.pe_each_iter = return_array[2]
                self.pe_each_iter += return_array[-1][2]
            if (self.want_bf == True):
                self.bf_each_iter = return_array[-2]
                self.bf_each_iter += return_array[-1][-2]

        else:
            # This is the mice case, where we have multiple return values.
            # We leave those as arrays.
            self.df_units_and_covars_matched = []
            self.groups_per_unit = []
            self.df_units_and_covars_matched = []
            self.units_per_group = []
            self.pe_each_iter = []
            self.bf_each_iter = []
            for return_val in return_array:
                self.df_units_and_covars_matched.append(return_val[0].drop(columns = ['weights']))
                self.groups_per_unit.append(return_val[0]['weights'])
                self.units_per_group.append(return_val[1])
                if self.want_pe:
                    self.pe_each_iter.append(return_val[2])
                if self.want_bf:
                    self.bf_each_iter.append(return_val[-1])


        return self.df_units_and_covars_matched


def _DAME(df, df_holdout, treatment_column_name='treated', weight_array=False,
          outcome_column_name='outcome', adaptive_weights='ridge', alpha=0.1,
          repeats=True, verbose=2, want_pe=False,
          early_stop_iterations=False, stop_unmatched_c=False,
          early_stop_un_c_frac=False, stop_unmatched_t=False,
          early_stop_un_t_frac=False, early_stop_pe=False,
          early_stop_pe_frac=0.01, want_bf=False, missing_indicator=np.nan,
          missing_data_replace=0, missing_holdout_replace=0,
          missing_holdout_imputations=10, missing_data_imputations=1):
    """ Accepts user input, validates, error-checks, calls DAME algorithm.

    Args:
        See description in MatchParent class above.

    Returns:
        return_df: df of units with the column values of their main matched
            group, with "*"s in place for the columns not in their MMG;
            includes a unit weights column which indicates the number of times
            each unit was matched
        MG_units: list of unit ids for every matched group
        pe_array: If want_pe is true, then the PE values of each match
        bf_array: If want_bf is true, then the BF values of each match

    Raises:
        Exception: An error occurred in the data_cleaning.py file.
    """

    df = data_cleaning.process_input_file(
        df, treatment_column_name, outcome_column_name)

    data_cleaning.check_parameters(adaptive_weights, df_holdout, df,
                                   alpha, False, weight_array)

    df, df_holdout, mice_on_match, mice_on_hold = data_cleaning.check_missings(
        df, df_holdout, missing_indicator, missing_data_replace,
        missing_holdout_replace, missing_holdout_imputations,
        missing_data_imputations, treatment_column_name, outcome_column_name,
        adaptive_weights)

    early_stops = data_cleaning.check_stops(
        stop_unmatched_c, early_stop_un_c_frac, stop_unmatched_t,
        early_stop_un_t_frac, early_stop_pe, early_stop_pe_frac,
        early_stop_iterations)

    if not mice_on_match:
        return dame_algorithm.algo1(
            df, treatment_column_name, weight_array, outcome_column_name,
            adaptive_weights, alpha, df_holdout, repeats, want_pe, verbose,
            want_bf, mice_on_hold, early_stops)

    # if the 'if' condition is not true, this would mean we need to run mice on
    # the matching data, which means that we have to run algo1 multiple times
    print("Warning: You have opted to run MICE on the matching dataset. "\
          "This is slow, and not recommended. We recommend that instead,"\
          " you run the algorithm and skip matching on missing data "\
              "points, with the parameter missing_data_replace=2.")

    # first we get the imputed datasets
    df_array = flame_dame_helpers.create_mice_dfs(df, mice_on_match,
                                                  outcome_column_name)
    return_array = []
    for i in range(len(df_array)):
        return_array.append(dame_algorithm.algo1(
            df_array[i], treatment_column_name, weight_array,
            outcome_column_name, adaptive_weights, alpha, df_holdout,
            repeats, want_pe, verbose, want_bf, mice_on_hold, early_stops))
    return return_array


def _FLAME(df, df_holdout, treatment_column_name='treated', weight_array=False,
           outcome_column_name='outcome', adaptive_weights='ridge', alpha=0.1,
           repeats=True, verbose=2, want_pe=False,
           early_stop_iterations=False, stop_unmatched_c=False,
           early_stop_un_c_frac=False, stop_unmatched_t=False,
           early_stop_un_t_frac=False, early_stop_pe=False,
           early_stop_pe_frac=0.01, want_bf=False,
           missing_indicator=np.nan,
           missing_data_replace=0, missing_holdout_replace=0,
           missing_holdout_imputations=10, missing_data_imputations=0,
           pre_dame=False, C=0.1):
    """ This function kicks off the FLAME algorithm.

    Args:
        See DAME above. The exeption is no weight_array, and the additional
        params below:

        pre_dame (int, False): Indicates whether to switch to dame and after
            int number of iterations.
        C (float, 0.1): The tradeoff between PE and BF in computing MQ


    Returns:
        See DAME above.
    """

    df = data_cleaning.process_input_file(
        df, treatment_column_name, outcome_column_name)

    data_cleaning.check_parameters(
        adaptive_weights, df_holdout, df, alpha, True, weight_array, C)

    df, df_holdout, mice_on_match, mice_on_hold = data_cleaning.check_missings(
        df, df_holdout, missing_indicator, missing_data_replace,
        missing_holdout_replace, missing_holdout_imputations,
        missing_data_imputations, treatment_column_name, outcome_column_name,
        adaptive_weights)

    early_stops = data_cleaning.check_stops(
        stop_unmatched_c, early_stop_un_c_frac, stop_unmatched_t,
        early_stop_un_t_frac, early_stop_pe, early_stop_pe_frac,
        early_stop_iterations)

    if not mice_on_match:
        return_array = flame_algorithm.flame_generic(
            df, treatment_column_name, weight_array, outcome_column_name,
            adaptive_weights, alpha, df_holdout, repeats, want_pe, verbose,
            want_bf, mice_on_hold, early_stops, pre_dame, C)

    else:
        # this would mean we need to run mice on the matching data, which means
        # that we have to run flame_generic multiple times
        if pre_dame:
            raise Exception("Invalid Input Error. At this time, we do not "\
                            "allow users to run the hybrid algorithm while "\
                            "running MICE on the matching dataset.")

        print("Warning: You have opted to run MICE on the matching dataset. "\
              "This is slow, and not recommended. We recommend that instead,"\
              " you run the algorithm and skip matching on missing data "\
              "points, with the parameter missing_data_replace=2.")

        # first we get the imputed datasets
        df_array = flame_dame_helpers.create_mice_dfs(df, mice_on_match,
                                                      outcome_column_name)

        return_array = []
        for i in range(len(df_array)):
            return_array.append(flame_algorithm.flame_generic(
                df_array[i], treatment_column_name, weight_array, outcome_column_name,
                adaptive_weights, alpha, df_holdout, repeats, want_pe, verbose,
                want_bf, mice_on_hold, early_stops, pre_dame, C))

    return return_array
